#include <gtk/gtk.h>
#include "guiutils.h"
#include "dspview.h"


/* Draw merge mode functions */
static void DSPViewDrawSet(guint32 *t, const guint32 *s);
static void DSPViewDrawSetAlpha(guint32 *t, const guint32 *s);
static void DSPViewDrawAlpha(guint32 *t, const guint32 *s);
static void DSPViewDrawAdd(guint32 *t, const guint32 *s);
static void DSPViewDrawSubtract(guint32 *t, const guint32 *s);
static void DSPViewDrawBlend(guint32 *t, const guint32 *s);
static void DSPViewDrawMultiply(guint32 *t, const guint32 *s);

/* Complex draw functions */
static void DSPViewDrawNovaDrawStar(
	guint32 *tar_buf, gint width, gint height,
	gint x, gint y,                 /* Center of star */
	gint swidth, gint sheight,      /* Size of star */
	guint32 v,
	gfloat intensity_coeff
);
static void DSPViewDrawNova(
	guint32 *tar_buf, guint32 *tar_buf2,
	gint width, gint height,
	guint32 v, gfloat fade_coeff,
	const guint8 *buf, gint len,
	gint channels
);
static void DSPViewDrawPlasma(
	guint32 *tar_buf, guint32 *tar_buf2,
	gint width, gint height,
	guint32 v, gfloat fade_coeff,
	const guint8 *buf, gint len,
	gint channels
);


/* Callbacks */
static void DSPViewRecreateBuffer(
	dspview_struct *dv, gint width, gint height
);
static gint DSPViewViewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);


/* Public */
dspview_struct *DSPViewNew(GtkWidget *parent, gint width, gint height);
void DSPViewSetBG(
	dspview_struct *dv,
	const guint8 *buf,
	gint width, gint height, gint bpl, gint bpp
);
void DSPViewSetChannels(
	dspview_struct *dv, gint channels
);
void DSPViewSetDisplayStyle(
	dspview_struct *dv, dspview_display_style style
);
void DSPViewSetMergeMode(
	dspview_struct *dv, dspview_merge_mode mode
);
void DSPViewSetDSPColor(
	dspview_struct *dv, guint32 c
);
void DSPViewSetFadeCoeff(
	dspview_struct *dv, gfloat v
);
void DSPViewDraw(
	dspview_struct *dv,
	const guint8 *buf, gint len
);
void DSPViewMap(dspview_struct *dv);
void DSPViewUnmap(dspview_struct *dv);
void DSPViewDelete(dspview_struct *dv);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Pixel merge modes functions, used by DSPViewDraw() to blit
 *	pixels.
 *
 *	The pointers t and s point to the target and source pixels
 *	(respectively).
 */
static void DSPViewDrawSet(guint32 *t, const guint32 *s)
{
	*t = *s;
}
/* Same as DSPViewDrawSet() except that it takes the source pixel's
 * alpha value into consideration, but does not modify the target
 * pixel's alpha value */
static void DSPViewDrawSetAlpha(guint32 *t, const guint32 *s)
{
	guint8		*t8 = (guint8 *)t;
	const guint8	*s8 = (guint8 *)s;
	gfloat		ac_s = (gfloat)s8[3] / (gfloat)0xff,
			ac_t = 1.0f - ac_s;

	*t = (guint32)(
	    ((guint32)((s8[0] * ac_s) + (t8[0] * ac_t)) << 0) |
	    ((guint32)((s8[1] * ac_s) + (t8[1] * ac_t)) << 8) |
	    ((guint32)((s8[2] * ac_s) + (t8[2] * ac_t)) << 16) |
	    ((guint32)t8[3] << 24)
	);

}
/* Only sets the target pixel's alpha value to match the source pixel's
 * alpha value */
static void DSPViewDrawAlpha(guint32 *t, const guint32 *s)
{
	const guint8 *s8 = (guint8 *)s;
	*t = (guint32)(
	    ((*t) & 0x00ffffff) |
	    ((guint32)s8[3] << 24)
	);

}
/* Adds source rgb value to the target pixel's value (does not warp),
 * the alpha value is not modified */
static void DSPViewDrawAdd(guint32 *t, const guint32 *s)
{
	guint8		*t8 = (guint8 *)t;
	const guint8	*s8 = (guint8 *)s;
	gfloat          ac_s = (gfloat)s8[3] / (gfloat)0xff;

	*t = (guint32)(
((guint32)MIN((gint)(s8[0] * ac_s) + (gint)t8[0], 0xff) << 0) |
((guint32)MIN((gint)(s8[1] * ac_s) + (gint)t8[1], 0xff) << 8) |
((guint32)MIN((gint)(s8[2] * ac_s) + (gint)t8[2], 0xff) << 16) |
	    ((guint32)t8[3] << 24)
	);
}
/* Subtracts source rgb value from the target pixel's value (does not
 * warp), the alpha value is not modified */
static void DSPViewDrawSubtract(guint32 *t, const guint32 *s)
{
	guint8		*t8 = (guint8 *)t;
	const guint8	*s8 = (guint8 *)s;
	gfloat          ac_s = (gfloat)s8[3] / (gfloat)0xff;

	*t = (guint32)(
((guint32)MAX((gint)t8[0] - (gint)(s8[0] * ac_s), 0x00) << 0) |
((guint32)MAX((gint)t8[1] - (gint)(s8[1] * ac_s), 0x00) << 8) |
((guint32)MAX((gint)t8[2] - (gint)(s8[2] * ac_s), 0x00) << 16) |
	    ((guint32)t8[3] << 24)
	);
}
/* Averages source rgb value with the target rgb value, the source
 * pixel's alpha value is taken into consideration (however the target
 * pixel's alpha value is not modified) */
static void DSPViewDrawBlend(guint32 *t, const guint32 *s)
{
	guint8          *t8 = (guint8 *)t;
	const guint8    *s8 = (guint8 *)s;
	gfloat          ac_s = (gfloat)s8[3] / (gfloat)0xff;

	*t = (guint32)(
	    ((((guint32)(s8[0] * ac_s) + (guint32)t8[0]) / 2) << 0) |
	    ((((guint32)(s8[1] * ac_s) + (guint32)t8[1]) / 2) << 8) |
	    ((((guint32)(s8[2] * ac_s) + (guint32)t8[2]) / 2) << 16) |
	    ((guint32)t8[3] << 24)
	);
}
/* Multiplies the source rgb value with the target's rgb value, the
 * source pixel's alpha value is taken into consideration (however the
 * target pixel's alpha value is not modified) */
static void DSPViewDrawMultiply(guint32 *t, const guint32 *s)
{
	guint8          *t8 = (guint8 *)t;
	const guint8    *s8 = (guint8 *)s;
	gfloat          ac_s = (gfloat)s8[3] / (gfloat)0xff;

	*t = (guint32)(
	    (((guint32)(s8[0] * ac_s) * (guint32)t8[0] / 0xff) << 0) |
	    (((guint32)(s8[1] * ac_s) * (guint32)t8[1] / 0xff) << 8) |
	    (((guint32)(s8[2] * ac_s) * (guint32)t8[2] / 0xff) << 16) |
	    ((guint32)t8[3] << 24)
	);
}

/*
 *	Used by DSPViewDrawNova() to additively draw one "star".
 *
 *	tar_buf, width, height, and intensity_coeff are assumed valid.
 */
static void DSPViewDrawNovaDrawStar(
	guint32 *tar_buf, gint width, gint height,
	gint x, gint y,			/* Center of star */
	gint swidth, gint sheight,	/* Size of star */
	guint32 v,
	gfloat intensity_coeff
)
{
	gint sx, sy, tx, ty;
	gint	half_swidth = swidth / 2,
		half_sheight = sheight / 2;
	gfloat sx_intensity_coeff, sy_intensity_coeff, t_coeff;
	gfloat transparency_coeff = (gfloat)((v & 0xff000000) >> 24) /
	    (gfloat)0xff;
	guint32 v2;


	if((swidth <= 0) || (sheight <= 0))
	    return;
	if((half_swidth <= 0) || (half_sheight <= 0))
	    return;

	/* Star (source) buffer totally not visible on target buffer? */
	if(((x + half_swidth) < 0) || ((x - half_swidth) >= width) ||
	   ((x + half_sheight) < 0) || ((y - half_sheight) >= height)
	)
	    return;

	/* Iterate through source y axis */
	for(sy = 0; sy < sheight; sy++)
	{
	    /* Calculate target y coordinate for this pixel */
	    ty = sy + y - half_sheight;
	    if((ty < 0) || (ty >= height))
		continue;

	    /* Calculate source y intensity coefficient */
	    if(sy < half_sheight)
		sy_intensity_coeff = 1.0f - (gfloat)sy / (gfloat)half_sheight;
	    else
		sy_intensity_coeff = (
		    (gfloat)(sy - half_sheight) / (gfloat)half_sheight
		);
	    if(sy_intensity_coeff > 1.0f)
		continue;

	    /* Iterate through source x axis */
	    for(sx = 0; sx < swidth; sx++)
	    {
		/* Calculate target x coordinate for this pixel */
		tx = sx + x - half_swidth;
		if((tx < 0) || (tx >= width))
		    continue;

		/* Calculate source x intensity coefficient */
		if(sx < half_swidth)
		    sx_intensity_coeff = 1.0f - (gfloat)sx / (gfloat)half_swidth;
		else
		    sx_intensity_coeff = (
			(gfloat)(sx - half_swidth) / (gfloat)half_swidth
		    );
		if(sx_intensity_coeff > 1.0f)
		    continue;

		/* Calculate target intensity coefficient based on the
		 * x, y, and limiting intensity coefficients.
		 */
		t_coeff = CLIP(
		    (1.0f - (sx_intensity_coeff * sy_intensity_coeff)) *
		    intensity_coeff,
		    0.0f, 1.0f
		);
		v2 = (
		    (v & 0x00ffffff) |
		    ((guint32)(0xff * t_coeff * transparency_coeff) << 24)
		);
		DSPViewDrawAdd(
		    &tar_buf[(ty * width) + tx], &v2
		);
	    }
	}
}
/*
 *	Draws Nova, a star jittering behind a "hazy atmosphere".
 *
 *	Uses tar_buf2 to record past additive values of the "star".  The
 *	tar_buf2 is darkened and then a new star is drawn at a position
 * 	determined by the "jittering" of the DSP values.  Then tar_buf2
 *	is added to tar_buf.
 */
static void DSPViewDrawNova(
	guint32 *tar_buf, guint32 *tar_buf2,
	gint width, gint height,
	guint32 v, gfloat fade_coeff,
	const guint8 *buf, gint len,
	gint channels
)
{
	if(TRUE)
	{
	    guint32 v2;
	    guint32 *ptr, *ptr2, *end;
	    gint buf_pos, buf_v, buf_v_last, dv;
	    gint dv_hist[4];
	    gfloat dv_hist_c[4];

	    /* Darken tar_buf2 */
	    v2 = (
		((guint32)(0xff * fade_coeff) << 0) |
		((guint32)(0xff * fade_coeff) << 8) |
		((guint32)(0xff * fade_coeff) << 16) |
		0xff000000
	    );
	    for(ptr = (guint32 *)tar_buf2,
		end = ptr + (width * height);
		ptr < end;
		ptr++
	    )
		DSPViewDrawSubtract(ptr, &v2);

	    /* Scan the buffer for the greatest change on this loop,
	     * then use that greatest change to create a new star.
	     *
	     * The change in positive or negative determines the
	     * y position and the earlier or later in the buffer
	     * determines the x position.  The average volume of the
	     * buffer determines the intensity.
	     */
	    buf_v_last = ((gint)buf[0] - (gint)0x80) * 2;
	    memset(dv_hist, 0x00, 4 * sizeof(gint));
	    for(buf_pos = 1; buf_pos < len; buf_pos++)
	    {
		/* Get new buffer value and change rate */
		buf_v = ((gint)buf[buf_pos] - (gint)0x80) * 2;
		dv = buf_v_last - buf_v;

		/* Update dv_hist, the change in buffer value history.
		 * Index values have the following meanings, for 0 and 1
		 * it means any positive or negative (respectively)
		 * changes that occure in the positive values, for 2
		 * and 3 it means any positive or negative (respectively)
		 * changes that occure in the negative values.
		 *
		 * 0 and 2 will always be positive, 1 and 3 will always
		 * be negative.
		 */
		if(buf_v < 0x00)
		{
		    /* Negative values, update dv_hist indices 2 and 3 */
		    if(dv > dv_hist[2])
			dv_hist[2] = dv;
		    else if(dv < dv_hist[3])
			dv_hist[3] = dv;
		}
		else
		{
		    /* Positive values, update dv_hist indices 0 and 1 */
		    if(dv > dv_hist[0])
			dv_hist[0] = dv;
		    else if(dv < dv_hist[1])
			dv_hist[1] = dv;
		}

		buf_v_last = buf_v;
	    }

	    /* Calculate delta change history values as coefficients */
	    dv_hist_c[0] = (gfloat)dv_hist[0] / (gfloat)0xff;
	    dv_hist_c[1] = (gfloat)dv_hist[1] / (gfloat)0xff;
	    dv_hist_c[2] = (gfloat)dv_hist[2] / (gfloat)0xff;
	    dv_hist_c[3] = (gfloat)dv_hist[3] / (gfloat)0xff;

	    /* Draw a new star */
	    DSPViewDrawNovaDrawStar(
		tar_buf2, width, height,
		(gint)(
		    (width / 2) + ((width / 2) *
		    ((dv_hist_c[0] > ABSOLUTE(dv_hist_c[1])) ?
			dv_hist_c[0] : dv_hist_c[1])
		    )
		),
		(gint)(
		    (height / 2) + ((height / 2) *
		    ((dv_hist_c[2] > ABSOLUTE(dv_hist_c[3])) ?
			dv_hist_c[2] : dv_hist_c[3])
		    )
		),
		(gint)(width * 2.0f),
		(gint)(height * 2.0f),
		v,
		((dv_hist_c[0] > 0.0f) && (dv_hist_c[2] > 0.0f)) ?
		    0.5f : 0.0f
	    );

	    /* Add tar_buf2 to tar_buf */
	    for(ptr = (guint32 *)tar_buf,
		ptr2 = (guint32 *)tar_buf2,
		end = ptr + (width * height);
		ptr < end;
		ptr++, ptr2++
	    )
		DSPViewDrawAdd(ptr, ptr2);
	}
}

/*
 *	Draws Plasma, a "globby" stream of plasma beads.
 *
 *	Uses tar_buf2 to record the graphics from the previous loop. The
 *	tar_buf2 is first darkened and then the new plasma is added to
 *	it.  Then tar_buf2 is added to tar_buf.
 */
static void DSPViewDrawPlasma(
	guint32 *tar_buf, guint32 *tar_buf2,
	gint width, gint height,
	guint32 v, gfloat fade_coeff,
	const guint8 *buf, gint len,
	gint channels
)
{
	if(TRUE)
	{
	    gint x, y, buf_pos;
	    gint dv, buf_v, last_buf_v;
	    gfloat v_c, a_c, t_y;
	    guint32 v2;
	    guint32 *ptr, *ptr2, *end;


	    /* Darken tar_buf2 */
	    v2 = (
		((guint32)(0x40 * fade_coeff) << 0) |
		((guint32)(0x40 * fade_coeff) << 8) |
		((guint32)(0x40 * fade_coeff) << 16) |
		0xff000000
	    );
	    for(ptr = (guint32 *)tar_buf2,
		end = ptr + (width * height);
		ptr < end;
		ptr++
	    )
		DSPViewDrawSubtract(ptr, &v2);

	    /* Initially reset last buffer value as -0xff to 0xff */
	    last_buf_v = ((gint)buf[0] - (gint)0x80) * 2;

	    /* Iterate through x coordinate while drawing each y axis
	     * "plasma strip".
	     */
	    for(x = 0; x < width; x++)
	    {
		/* Get buffer value that matches this x coordinate
		 * position plus 1.  Get buf_v as a value from -0xff to
		 * 0xff.
		 */
		buf_pos = (x * len / width) + 1;
		if((buf_pos >= 0) && (buf_pos < len))
		    buf_v = ((gint)buf[buf_pos] - (gint)0x80) * 2;
		else
		    buf_v = last_buf_v;

		/* Calculate buffer value change dv and apply the fade
		 * coefficient to it.  Then update buf_v to reflect the
		 * "dampened" change.
		 */
		dv = (buf_v - last_buf_v) * fade_coeff;
		buf_v = last_buf_v + dv;

		/* Calculate buffer value coeff from -1.0 to 1.0 */
		v_c = (gfloat)buf_v / (gfloat)0xff;

		/* Draw this y axis "plasma strip" */
		for(y = 0; y < height; y++)
		{
		    t_y = ((gfloat)y / (gfloat)height) - 0.5f;

		    /* Calculate alpha value coeff from -1.0 to 1.0.
		     *
		     * The equation used is:
		     *
		     * c = a * (1 - (w * x^2))
		     *
		     * Where a is the amplitude (makes the parabola
		     * shallow or steep) and w is the wideness of the
		     * parabola (4.5 works best, 4.0 makes it a bit
		     * wider).
		     */
		    if(v_c < 0.0f)
			a_c = MIN(
			    v_c * (1.0f - (4.0f * t_y * t_y)),
			    0.0f
			);
		    else
			a_c = MAX(
			    v_c * (1.0f - (4.0f * t_y * t_y)),
			    0.0f
			);
		    if(a_c == 0.0f)
			continue;

		    v2 = (guint32)(
(v & 0x00ffffff) |
((guint32)((gfloat)0xff * ABSOLUTE(a_c)) << 24)
		    );

		    DSPViewDrawAdd(
			&tar_buf2[(y * width) + x], &v2
		    );
		}

		last_buf_v = buf_v;
	    }

	    /* Add tar_buf2 to tar_buf */
	    for(ptr = (guint32 *)tar_buf,
		ptr2 = (guint32 *)tar_buf2,
		end = ptr + (width * height);
		ptr < end;
		ptr++, ptr2++
	    )
		DSPViewDrawAdd(ptr, ptr2);
	}

}

/*
 *	Reallocates the drawing buffer to the given size in pixels.
 */
static void DSPViewRecreateBuffer(
	dspview_struct *dv, gint width, gint height
)
{
	if(dv == NULL)
	    return;

	g_free(dv->buf);
	dv->buf = NULL;

	g_free(dv->dsp_buf);
	dv->dsp_buf = NULL;

	if((width > 0) && (height > 0))
	{
	    dv->bpp = 4;
	    dv->bpl = width * dv->bpp;
	    dv->width = width;
	    dv->height = height;
	    dv->buf = (guint8 *)g_malloc(
		dv->bpl * dv->height * sizeof(guint8)
	    );

	    /* Allocate dsp_buf to match size of main buffer */
	    dv->dsp_buf = (guint8 *)g_malloc(
		dv->bpl * dv->height * sizeof(guint8)
	    );
	    /* Need to reset dsp_buf after allocation */
	    if(dv->dsp_buf != NULL)
	    {
		guint32 *ptr = (guint32 *)dv->dsp_buf,
			*end = ptr + (dv->width * dv->height);
		while(ptr < end)
		    *ptr++ = 0xff000000;
	    }
	}
}

/*
 *	View GtkDrawingArea event signal callback.
 */
static gint DSPViewViewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	dspview_struct *dv = DSPVIEW(data);
	if((widget == NULL) || (event == NULL) || (dv == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    DSPViewRecreateBuffer(
		dv, configure->width, configure->height
	    );
	    DSPViewDraw(dv, NULL, 0);
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    DSPViewDraw(dv, NULL, 0);
	    status = TRUE;
	    break;
	}

	return(status);
}


/*
 *	Creates a new DSP view and parents it to the given parent.
 */
dspview_struct *DSPViewNew(GtkWidget *parent, gint width, gint height)
{
	GtkWidget *w, *parent2;
	dspview_struct *dv = DSPVIEW(
	    g_malloc0(sizeof(dspview_struct))
	);
	if(dv == NULL)
	    return(NULL);

	/* Reset values */
	dv->map_state = FALSE;
	dv->channels = 1;
	dv->display_style = DSPVIEW_DISPLAY_STYLE_WAVE;
	dv->merge_mode = DSPVIEW_MERGE_MODE_SET;
	dv->dsp_color = DSPVIEW_DEF_DSP_COLOR;	/* 0xAABBGGRR */
	dv->fade_coeff = DSPVIEW_DEF_FADE_COEFF;
	dv->buf = NULL;
	dv->dsp_buf = NULL;
	dv->bg_buf = NULL;
	memset(
	    dv->analyzer_history_buf, 0x00,
	    DSPVIEW_TOTAL_ANALYZER_BLOCKS * sizeof(guint8)
	);

	/* Create toplevel, a GtkFrame */
	dv->toplevel = parent2 = w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_widget_set_usize(w, width, height);
	if(parent != NULL)
	{
	    if(GTK_IS_BOX(parent))
		gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    else if(GTK_IS_CONTAINER(parent))
		gtk_container_add(GTK_CONTAINER(parent), w);
	}

	/* View GtkDrawingArea */
	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
	dv->view_da = w = gtk_drawing_area_new();
	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();
	gtk_widget_add_events(
	    w,
	    GDK_STRUCTURE_MASK | GDK_VISIBILITY_NOTIFY_MASK |
	    GDK_EXPOSURE_MASK | GDK_FOCUS_CHANGE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(DSPViewViewEventCB), dv
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "visibility_notify_event",
	    GTK_SIGNAL_FUNC(DSPViewViewEventCB), dv
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(DSPViewViewEventCB), dv
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);

	return(dv);
}

/*
 *	Sets the background image.
 *
 *	bpp must be 4 and the format type must be RGBA.
 */
void DSPViewSetBG(
	dspview_struct *dv,
	const guint8 *buf,
	gint width, gint height, gint bpl, gint bpp
)
{
	if(dv == NULL)
	    return;

	dv->bg_buf = buf;
	dv->bg_width = width;
	dv->bg_height = height;
	dv->bg_bpl = bpl;
	dv->bg_bpp = bpp;
}

/*
 *	Sets the number of channels on the DSPView, must be 1, 2, or 4.
 */
void DSPViewSetChannels(
	dspview_struct *dv, gint channels
)
{
	if(dv == NULL)
	    return;

	dv->channels = channels;
}

/*
 *	Sets the DSPView display style.
 */
void DSPViewSetDisplayStyle(
	dspview_struct *dv, dspview_display_style style
)
{
	if(dv == NULL)
	    return;

	if(dv->display_style != style)
	{
	    /* Always clear DSP buffer when changing display styles */
	    if(dv->dsp_buf != NULL)
	    {
		guint32	*ptr = (guint32 *)dv->dsp_buf,
			*end = ptr + (dv->width * dv->height);
		while(ptr < end)
		    *ptr++ = 0xff000000;
	    }


	    if(style == DSPVIEW_DISPLAY_STYLE_ANALYZER)
	    {
		/* Need to reset analyzer history buffer when switching
		 * to it.
		 */
		memset(
		    dv->analyzer_history_buf, 0x00,
		    DSPVIEW_TOTAL_ANALYZER_BLOCKS * sizeof(guint8)
		);
	    }
	}

	/* Set new display style */
	dv->display_style = style;
}

/*
 *      Sets the DSPView merge mode.
 */
void DSPViewSetMergeMode(
	dspview_struct *dv, dspview_merge_mode mode
)
{
	if(dv == NULL)
	    return;

	dv->merge_mode = mode;	    
}

/*
 *	Sets the DSPView color in 0xAABBGGRR format.
 */
void DSPViewSetDSPColor(
	dspview_struct *dv, guint32 c
)
{
	if(dv == NULL)
	    return;

	dv->dsp_color = c;
}

/*
 *	Sets the fade coeff.
 */
void DSPViewSetFadeCoeff(
	dspview_struct *dv, gfloat v
)
{
	if(dv == NULL)
	    return;

	dv->fade_coeff = CLIP(v, 0.0f, 1.0f);
}

/*
 *	Redraws the DSP view.
 */
void DSPViewDraw(
	dspview_struct *dv,
	const guint8 *buf, gint len
)
{
	gint state, channels;
	GdkWindow *window;
	GtkWidget *w;
	GtkStyle *style;
	guint8 *img_buf;
	gint width, height, bpl, bpp;
	void (*bg_set_func)(guint32 *, const guint32 *) = DSPViewDrawSet;
	void (*set_func)(guint32 *, const guint32 *) = DSPViewDrawSet;
	if(dv == NULL)
	    return;

	w = dv->view_da;
	if((w != NULL) ? !GTK_WIDGET_VISIBLE(w) : TRUE)
	    return;

	channels = dv->channels;
	state = GTK_WIDGET_STATE(w);
	window = w->window;
	style = gtk_widget_get_style(w);
	if((window == NULL) || (style == NULL))
	    return;

	img_buf = dv->buf;
	width = dv->width;
	height = dv->height;
	bpl = dv->bpl;
	bpp = dv->bpp;

	if((img_buf == NULL) || (width <= 0) || (height <= 0))
	    return;

	/* Set the DSP merge mode function as set_func depending on the
	 * current merge_mode so each blit loop uses set_func to merge
	 * the source pixel to the target pixel.
	 */
	switch(dv->merge_mode)
	{
	  case DSPVIEW_MERGE_MODE_SET:
	    set_func = DSPViewDrawSet;
	    break;
	  case DSPVIEW_MERGE_MODE_SET_ALPHA:
	    set_func = DSPViewDrawSetAlpha;
	    break;
	  case DSPVIEW_MERGE_MODE_ADD:
	    set_func = DSPViewDrawAdd;
	    break;
	  case DSPVIEW_MERGE_MODE_SUBTRACT:
	    set_func = DSPViewDrawSubtract;
	    break;
	  case DSPVIEW_MERGE_MODE_BLEND:
	    set_func = DSPViewDrawBlend;
	    break;
	  case DSPVIEW_MERGE_MODE_MULTIPLY:
	    set_func = DSPViewDrawMultiply;
	    break;
	}

	/* Draw background, thus clearing the back buffer */
	if(TRUE)
	{
	    gint x, y;

	    if((dv->bg_buf != NULL) && (dv->bg_width > 0) &&
	       (dv->bg_height > 0)
	    )
	    {
		const guint32 *src_buf = (const guint32 *)dv->bg_buf;
		guint32 *tar_buf = (guint32 *)img_buf;
		gint	bg_width = dv->bg_width,
			bg_height = dv->bg_height;
		const guint8 *cptr8;
		guint32 v2, g;

		if(state == GTK_STATE_INSENSITIVE)
		{
		    for(y = 0; y < height; y++)
		    {
			for(x = 0; x < width; x++)
			{
			    /* Get source pixel and make it grey */
			    cptr8 = (const guint8 *)&src_buf[
		((y % bg_height) * bg_width) + (x % bg_width)
			    ];
			    g = ((gint)cptr8[0] + (gint)cptr8[1] + (gint)cptr8[2]) / 3;
			    v2 = (guint32)(
				(g << 0) |
				(g << 8) |
				(g << 16) |
				((guint32)cptr8[3] << 24)
			    );
			    bg_set_func(
				&tar_buf[(y * width) + x], &v2
			    );
			}
		    }
		}
		else
		{
		    for(y = 0; y < height; y++)
		    {
			for(x = 0; x < width; x++)
			    bg_set_func(
				&tar_buf[(y * width) + x],
				&src_buf[
		((y % bg_height) * bg_width) + (x % bg_width)
				]
			    );
		    }
		}
	    }
	    else
	    {
		guint32 *tar_buf = (guint32 *)img_buf;
		guint32 v = 0xff000000;

		for(y = 0; y < height; y++)
		{
		    for(x = 0; x < width; x++)
			bg_set_func(
			    &tar_buf[(y * width) + x], &v
			);
		}
	    }
	}


	/* Draw the given buffer depending on the display style */
	switch(dv->display_style)
	{
	  case DSPVIEW_DISPLAY_STYLE_NOVA:
	    if((buf != NULL) && (len > 0) && (dv->dsp_buf != NULL) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		DSPViewDrawNova(
		    (guint32 *)img_buf, (guint32 *)dv->dsp_buf,
		    width, height, dv->dsp_color, dv->fade_coeff,
		    buf, len, channels
		);
	    }
	    break;

	  case DSPVIEW_DISPLAY_STYLE_PLASMA:
	    if((buf != NULL) && (len > 0) && (dv->dsp_buf != NULL) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		DSPViewDrawPlasma(
		    (guint32 *)img_buf, (guint32 *)dv->dsp_buf,
		    width, height, dv->dsp_color, dv->fade_coeff,
		    buf, len, channels
		);
	    }
	    break;

	  case DSPVIEW_DISPLAY_STYLE_LIGHTENING:
	    if((buf != NULL) && (len > 0) && (dv->dsp_buf != NULL) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		gint x, y, prev_y, dy, buf_pos;
		guint32 *tar_buf = (guint32 *)img_buf;
		guint32 *tar_buf2, *end;
		const guint8 *src_buf = (const guint8 *)buf;
		guint32 v = dv->dsp_color;
		guint32 subtract_v = 0xff0a0a0a;
		gfloat fade_coeff = CLIP(dv->fade_coeff, 0.0, 1.0);

		/* Darken each pixel on the buffer tar_buf2, this is to
		 * make the last lightening "fade away".
		 */
		for(tar_buf2 = (guint32 *)dv->dsp_buf,
		    end = tar_buf2 + (width * height);
		    tar_buf2 < end;
		    tar_buf2++
		)
		    DSPViewDrawSubtract(tar_buf2, &subtract_v);

		/* Add new lightening to tar_buf2 */
		tar_buf2 = (guint32 *)dv->dsp_buf;
		prev_y = (gint)src_buf[0] * height / (gint)0xff;
		for(x = 0; x < width; x++)
		{
		    buf_pos = x * len / width;
		    if((buf_pos < 0) || (buf_pos >= len))
			break;

		    y = (gint)src_buf[buf_pos] * height / (gint)0xff;
		    dy = (y - prev_y) * fade_coeff;
		    y = prev_y + dy;

		    if((y >= 0) && (y < height))
			DSPViewDrawAdd(
			    &tar_buf2[(y * width) + x], &v
			);

		    prev_y = y;
		}

		/* Add tar_buf2 to tar_buf */
		while(tar_buf2 < end)
		{
		    DSPViewDrawAdd(tar_buf, tar_buf2);
		    tar_buf++; tar_buf2++;
		}
	    }
	    break;

	  case DSPVIEW_DISPLAY_STYLE_ANALYZER:
	    if((buf != NULL) && (len > 0) && (dv->dsp_buf != NULL) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		gint i, x, y, buf_pos;
		guint32 *ptr, *ptr2, *end;
		guint32 *tar_buf = (guint32 *)img_buf;
		guint32 *tar_buf2 = (guint32 *)dv->dsp_buf;
		const guint8 *src_buf = (const guint8 *)buf;
		guint32 v = dv->dsp_color, v2;
		guint8 analyzer_buf[DSPVIEW_TOTAL_ANALYZER_BLOCKS];
		gfloat fade_coeff = CLIP(dv->fade_coeff, 0.0, 1.0);


		/* Darken all pixels on tar_buf2 */
		v2 = (
		    ((guint32)(0x80 * fade_coeff) << 0) |
		    ((guint32)(0x80 * fade_coeff) << 8) |
		    ((guint32)(0x80 * fade_coeff) << 16) |
		    0xff000000
		);
		for(ptr = tar_buf2,
		    end = ptr + (width * height);
		    ptr < end;
		    ptr++
		)
		    DSPViewDrawSubtract(ptr, &v2);


		/* Draw old position bar from bottom to position */
		for(x = 0; x < width; x++)
		{
		    i = x * DSPVIEW_TOTAL_ANALYZER_BLOCKS / width;
		    if((i < 0) || (i >= DSPVIEW_TOTAL_ANALYZER_BLOCKS))
			break;

		    for(y = MAX(height - (
			(gint)dv->analyzer_history_buf[i] * height /
			(gint)0xff) - 1, 0);
			y < height;
			y++
		    )
			DSPViewDrawAdd(
			    &tar_buf2[(y * width) + x], &v
			);
		}

		/* Get value from DSP buffer and copy to new
		 * analyzer_buf, then adjust the history analyzer
		 * analyzer_history_buf to "come close" to the new
		 * value.
		 */
		for(i = 0; i < DSPVIEW_TOTAL_ANALYZER_BLOCKS; i++)
		{
		    buf_pos = i * len / DSPVIEW_TOTAL_ANALYZER_BLOCKS;
		    if((buf_pos < 0) || (buf_pos >= len))
			break;

		    /* Copy value from DSP buffer to new analyzer
		     * buffer.
		     */
		    analyzer_buf[i] = (guint8)(
			ABSOLUTE((gint)src_buf[buf_pos] - (gint)0x80) * 2
		    );

		    /* Update the last analyzer bar position to "come
		     * close" to the new position.  Note that we use
		     * the fade_coeff on delta change value y to adjust
		     * how much the new analyzer value changes.
		     *
		     * Note that fade_coeff usually is not enough, so we
		     * dampen it further with a * 0.4.
		     */
		    x = (gint)dv->analyzer_history_buf[i];
		    y = ((gint)analyzer_buf[i] - (gint)x) * fade_coeff * 0.4;
		    if(((y < 0) && (x > 0x00)) ||
		       ((y > 0) && (x < 0xff))
		    )
			dv->analyzer_history_buf[i] += y;
		}

		/* Draw new position bar */
		for(x = 0; x < width; x++)
		{
		    i = x * DSPVIEW_TOTAL_ANALYZER_BLOCKS / width;
		    if((i < 0) || (i >= DSPVIEW_TOTAL_ANALYZER_BLOCKS))
			break;

		    y = height - (
			(gint)analyzer_buf[i] * height /
			(gint)0xff
		    ) - 1;

		    /* Draw new position bar only if it is "higher"
		     * than the previous position bar.
		     */
		    if((y >= 0) && (y < height) &&
		       (analyzer_buf[i] > dv->analyzer_history_buf[i])
		    )
			DSPViewDrawAdd(
			    &tar_buf2[(y * width) + x], &v
			);
		}

		/* Copy tar_buf2 to tar_buf */
		for(ptr = tar_buf,
		    ptr2 = tar_buf2,
		    end = ptr + (width * height);
		    ptr < end;
		    ptr++, ptr2++
		)
		    DSPViewDrawAdd(ptr, ptr2);

	    }
	    break;

	  case DSPVIEW_DISPLAY_STYLE_SGRAPH:
	    if((buf != NULL) && (len > 0) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		gint x, y, buf_pos;
		guint32 *tar_buf = (guint32 *)img_buf;
		const guint8 *src_buf = (const guint8 *)buf;
		guint32 v = dv->dsp_color;

		for(x = 0; x < width; x++)
		{
		    buf_pos = x * len / width;
		    if((buf_pos < 0) || (buf_pos >= len))
			break;

		    y = (gint)src_buf[buf_pos] * height / (gint)0xff;
		    if((y >= 0) && (y < height))
			set_func(
			    &tar_buf[(y * width) + x], &v
			);
		}
	    }
	    break;

	  case DSPVIEW_DISPLAY_STYLE_WAVE:
	    if((buf != NULL) && (len > 0) &&
	       (state != GTK_STATE_INSENSITIVE)
	    )
	    {
		gint x, y, dy, y_end, buf_pos;
		guint32 *tar_buf = (guint32 *)img_buf;
		const guint8 *src_buf = (const guint8 *)buf;
		guint32 v = dv->dsp_color;

		for(x = 0; x < width; x++)
		{
		    buf_pos = x * len / width;
		    if((buf_pos < 0) || (buf_pos >= len))
			break;

		    dy = ABSOLUTE((gint)src_buf[buf_pos] - (gint)0x80) * 2 *
			height / (gint)0xff;

		    for(y = MAX((height / 2) - dy, 0),
			y_end = MIN((height / 2) + dy + 1, height);
			y < y_end;
			y++
		    )
			set_func(
			    &tar_buf[(y * width) + x], &v
			);
		}
	    }
	    break;
	}


	/* Send view buffer to window */
	gdk_draw_rgb_32_image(
	    (GdkDrawable *)window, style->fg_gc[state],
	    0, 0,
	    width, height,
	    GDK_RGB_DITHER_NORMAL,
	    (guchar *)img_buf,
	    bpl
	);
}

/*
 *	Maps the DSPView.
 */
void DSPViewMap(dspview_struct *dv)
{
	GtkWidget *w = (dv != NULL) ? dv->toplevel : NULL;
	if(w == NULL)
	    return;

	gtk_widget_show(w);
	dv->map_state = TRUE;
}

/*
 *      Unmaps the DSPView.
 */
void DSPViewUnmap(dspview_struct *dv)
{
	GtkWidget *w = (dv != NULL) ? dv->toplevel : NULL; 
	if(w == NULL) 
	    return;  

	gtk_widget_hide(w);
	dv->map_state = FALSE;
}

/*
 *	Deletes the DSPView.
 */
void DSPViewDelete(dspview_struct *dv)
{
	if(dv == NULL)
	    return;

	GTK_WIDGET_DESTROY(dv->view_da)
	GTK_WIDGET_DESTROY(dv->toplevel)

	g_free(dv->buf);
	dv->buf = NULL;
	g_free(dv->dsp_buf);
	dv->dsp_buf = NULL;

	g_free(dv);
}
