#include <stdio.h>
#include <unistd.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <Imlib.h>

#include <gtk/gtk.h>
#include <gdk/gdkrgb.h>
#include <gdk/gdkx.h>

#include "../include/imlibio.h"

#include "guiutils.h"
#include "cdialog.h"
#include "csd.h"
#include "dspview.h"
#include "conbar.h"

#include "images/icon_yconsole_48x48.xpm"
#include "images/dspview_bg1.xpm"


/*
 *	Window structure.
 */
typedef struct {

	GtkWidget *toplevel;
	connection_bar_struct *conbar;
	YConnection *con;

	guint8 *yshm_buf;
	gint yshm_buf_len, yshm_id;

	gchar *audio_mode_name;
	gint sample_rate;
	gint sample_size;
	gint channels;
	glong cycle_us;
	gfloat cycle_load;
	gint num_fragments;
	gint fragment_size;
	gboolean shellout;

	guint toid, toid2;

	guint8 *bg_img;
	gint bg_img_width, bg_img_height, bg_img_bpl;

} win_struct;


gpointer imlib_handle;


static void do_connect(win_struct *win, const gchar *s);
static void do_disconnect(win_struct *win, gboolean force_shutdown);

static gint delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);

static void connect_cb(
        GtkWidget *widget, const gchar *s, gpointer data
);
static void yserver_list_cb(GtkWidget *widget, gpointer data);
static void close_cb(GtkWidget *widget, gpointer data);
static void new_console_cb(GtkWidget *widget, gpointer data);
static void new_mixer_cb(GtkWidget *widget, gpointer data);
static void new_audio_cb(GtkWidget *widget, gpointer data);
static void new_player_cb(GtkWidget *widget, gpointer data);
static void options_cb(GtkWidget *widget, gpointer data);
static void refresh_cb(GtkWidget *widget, gpointer data);
static gboolean shellout_cb(
        GtkWidget *widget, gboolean shellout, gpointer data
);
static void shutdown_cb(GtkWidget *widget, gpointer data);
static gint timeout2_cb(gpointer data);
static gint timeout_cb(gpointer data);


/*
 *	Procedure to open connection to Y server.
 */
static void do_connect(win_struct *win, const gchar *s)
{
	YConnection *con;


        if((s == NULL) || (win == NULL))
            return;

	/* Disconnect from Y server as needed. */
	do_disconnect(win, FALSE);

	/* Connect to Y server. */
	win->con = con = YOpenConnection(
            NULL, s
        );
        if(con != NULL)
        {
            /* Connected to Y server, now get some initial values. */
            YEventAudioStats stat_buf;
	    dspview_struct *dv = ConBarGetDSPView(win->conbar);

	    /* Update connection bar connected state. */
	    ConBarSetConnectState(win->conbar, TRUE);

            /* YSHM sound buffer pointer. */
            win->yshm_buf = (guint8 *)YSHMOpenSound(
                con, &win->yshm_buf_len, &win->yshm_id
            );

            /* Audio stats. */
            if(!YGetAudioStats(con, &stat_buf))
            {
		g_free(win->audio_mode_name);
		win->audio_mode_name = g_strdup(stat_buf.mode_name);
                win->sample_rate = stat_buf.sample_rate;
                win->sample_size = stat_buf.sample_size;
                win->channels = stat_buf.channels;
		win->cycle_us = stat_buf.cycle_us;
		win->cycle_load = 0.0;
		win->num_fragments = stat_buf.num_fragments;
		win->fragment_size = stat_buf.fragment_size;
		win->shellout = FALSE;
		ConBarSetAudio(
		    win->conbar, win->audio_mode_name,
                    win->sample_rate,
                    win->sample_size,
                    win->channels,
                    win->cycle_us,
                    win->cycle_load,
                    win->num_fragments,
                    win->fragment_size,
                    win->shellout
		);
		ConBarSetShellOut(win->conbar, win->shellout);
                DSPViewSetChannels(dv, win->channels);
            }

	    /* Record location history on connection bar. */
	    ConBarSetLocation(win->conbar, s, FALSE, TRUE);
	}
}

/*
 *	Procedure to close connection to Y server.
 */
static void do_disconnect(win_struct *win, gboolean force_shutdown)
{
	dspview_struct *dv;


	if(win == NULL)
	    return;

	dv = ConBarGetDSPView(win->conbar);

	/* Update connection bar connected state. */
	ConBarSetConnectState(win->conbar, FALSE);

	/* Close YSHM sound buffer. */
	YSHMCloseSound(win->con, win->yshm_buf);
	win->yshm_buf = NULL;
	win->yshm_buf_len = 0;
	win->yshm_id = -1;

	/* Close connection or shutdown? */
	if(force_shutdown)
	    YShutdownServer(win->con);
	else
	    YCloseConnection(win->con, False);
	win->con = NULL;

	g_free(win->audio_mode_name);
	win->audio_mode_name = NULL;


	/* Update audio values on connection bar. */
	g_free(win->audio_mode_name);
	win->audio_mode_name = NULL;
	win->sample_rate = 0;
	win->sample_size = 0;
	win->channels = 0;
	win->cycle_us = 0;
	win->cycle_load = 0.0;
	win->num_fragments = 0;
	win->fragment_size = 0;
	win->shellout = FALSE;
	ConBarSetAudio(
	    win->conbar, win->audio_mode_name,
	    win->sample_rate,
            win->sample_size,
            win->channels,
            win->cycle_us,
            win->cycle_load,
            win->num_fragments,
            win->fragment_size,
            win->shellout
        );
	DSPViewSetChannels(dv, win->channels);
}


/*
 *      Window toplevel "delete_event" signal callback.
 */
static gint delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gtk_main_quit();
        return(TRUE);
}

/*
 *	Connect callback.
 */
static void connect_cb(
	GtkWidget *widget, const gchar *s, gpointer data
)
{
	gchar *strptr;
	gchar *s2 = (s != NULL) ? g_strdup(s) : NULL;

	/* Empty string? Then use default location. */
	if(*s2 == '\0')
	{
	    g_free(s2);
	    s2 = g_strdup("127.0.0.1:9433");
	}

	/* No port number? Then append it. */
	strptr = strchr(s2, ':');
	if(strptr == NULL)
	{
	    strptr = g_strdup_printf("%s:9433", s2);
	    g_free(s2);
	    s2 = strptr;
	}

	/* Do connect. */
	do_connect((win_struct *)data, s2);

	/* Deallocate coppied location. */
	g_free(s2);
}

/*
 *	Y server list callback.
 */
static void yserver_list_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *	Close callback.
 */
static void close_cb(GtkWidget *widget, gpointer data)
{
	delete_cb(widget, NULL, data);
}

/*
 *	New console callback.
 */
static void new_console_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *      New audio callback.
 */
static void new_audio_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *      New mixer callback.
 */
static void new_mixer_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *      New player callback.
 */
static void new_player_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *	Options callback.
 */
static void options_cb(GtkWidget *widget, gpointer data)
{

}

/*
 *	Refresh callback.
 */
static void refresh_cb(GtkWidget *widget, gpointer data)
{
	gchar *s;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return;

	do_disconnect(win, FALSE);
	s = ConBarGetLocation(win->conbar);
	s = (s != NULL) ? g_strdup(s) : NULL;
	do_connect(win, s);
	g_free(s);
}

/*
 *	Shellout callback.
 *
 *	Returns actual shellout state.
 */
static gboolean shellout_cb(
	GtkWidget *widget, gboolean shellout, gpointer data
)
{
	YConnection *con;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return(win->shellout);

	con = win->con;
	if(con == NULL)
	    return(win->shellout);

	if(shellout == win->shellout)
	    return(win->shellout);

	if(shellout)
	{
	    if(YSetAudioModeValues(
		con,
		8,	/* Sample size in bits. */
		1,	/* Number of channels. */
		0,	/* Sample rate set to 0 implies shellout. */
		0,	/* Direction, 0 = play, 1 = record. */
		True,
		2,
		1024
	    ))
		win->shellout = FALSE;
	    else
		win->shellout = TRUE;
	}
	else
	{
	    if(YChangeAudioModePreset(
		con,
		(win->audio_mode_name != NULL) ?
		    win->audio_mode_name : "Default"
	    ))
		win->shellout = TRUE;
            else
                win->shellout = FALSE;
	}

	ConBarSetAudio(
	    win->conbar, win->audio_mode_name,
	    win->sample_rate,
            win->sample_size,
            win->channels,
            win->cycle_us,
            win->cycle_load,
            win->num_fragments,
            win->fragment_size,
	    win->shellout
	);

	return(win->shellout);
}

/*
 *	Shutdown callback.
 */
static void shutdown_cb(GtkWidget *widget, gpointer data)
{
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return;

	do_disconnect(win, TRUE);
}

#define QUIT_GTK_MAIN_RETURN    \
{ \
 win->toid = (guint)-1; \
 win->toid2 = (guint)-1; \
 gtk_main_quit(); \
 return(FALSE); \
}
/*
 *      Main timeout callback.
 */
static gint timeout_cb(gpointer data)
{
        YConnection *con;
        YEvent yevent;
        dspview_struct *dv;
	connection_bar_struct *cb;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return(FALSE);

	cb = win->conbar;
        if(cb == NULL)
            QUIT_GTK_MAIN_RETURN

        dv = ConBarGetDSPView(cb);
        if(dv == NULL)
            QUIT_GTK_MAIN_RETURN

        /* Manage Y events. */
	con = win->con;
        while((con != NULL) && (YGetNextEvent(con, &yevent, False) > 0))
        {
            switch(yevent.type)
            {
              case YAudioChange:
                /* Need to reopen YSHM sound buffer when Audio
                 * mode changes.
                 */
                YSHMCloseSound(con, win->yshm_buf);
                win->yshm_buf = (guint8 *)YSHMOpenSound(
                    con, &win->yshm_buf_len, &win->yshm_id
                );
		/* Update Audio values on conbar and DSPView. */
		if(yevent.audio.preset)
		{
                    YEventAudioStats stat_buf;

		    g_free(win->audio_mode_name);
		    win->audio_mode_name = g_strdup(yevent.audio.mode_name);

                    if(!YGetAudioStats(con, &stat_buf))
                    {
			g_free(win->audio_mode_name);
			win->audio_mode_name = g_strdup(stat_buf.mode_name);
                        win->sample_rate = stat_buf.sample_rate;
                        win->sample_size = stat_buf.sample_size;
                        win->channels = stat_buf.channels;
                        win->cycle_us = stat_buf.cycle_us;
                        win->num_fragments = stat_buf.num_fragments;
                        win->fragment_size = stat_buf.fragment_size;
                        win->shellout = FALSE;
                    }
		}
		else
		{
		    YEventAudioChange *audio = &yevent.audio;
		    win->sample_rate = audio->sample_rate;
                    win->sample_size = audio->sample_size;
                    win->channels = audio->channels;
                    win->num_fragments = audio->num_fragments;
                    win->fragment_size = audio->fragment_size_bytes;
		    win->shellout = FALSE;
		}
		ConBarSetAudio(
		    cb, win->audio_mode_name,
                    win->sample_rate,
                    win->sample_size,
                    win->channels,
                    win->cycle_us,
                    win->cycle_load,
                    win->num_fragments,
                    win->fragment_size,
                    win->shellout
		);
		ConBarSetShellOut(cb, win->shellout);
		DSPViewSetChannels(dv, yevent.audio.channels);
                break;

	      case YCycleChange:
		/* Update audio values. */
		win->cycle_us = yevent.cycle.cycle_us;
		ConBarSetAudio(
                    cb, win->audio_mode_name,
                    win->sample_rate,
                    win->sample_size,
                    win->channels,
                    win->cycle_us,
		    win->cycle_load,
		    win->num_fragments,
		    win->fragment_size,
		    win->shellout
                );
		break;

              case YDisconnect: case YShutdown:
		do_disconnect(win, FALSE);
		con = NULL;
                break;
            }
        }

        /* Redraw DSP view. */
	ConBarDraw(cb, win->yshm_buf, win->yshm_buf_len);

        return(TRUE);
}

/*
 *	Secondary timeout callback, called every 500 ms.
 */
static gint timeout2_cb(gpointer data)
{
        YConnection *con;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return(FALSE);

        con = win->con;
	if(con != NULL)
	{
	    YEventServerStats stat_buf;

	    if(!YGetServerStats(con, &stat_buf))
	    {
		win->cycle_load = stat_buf.cycle_load;

                ConBarSetAudio(
                    win->conbar, win->audio_mode_name,
                    win->sample_rate,
                    win->sample_size,
                    win->channels,
                    win->cycle_us,
                    win->cycle_load,
                    win->num_fragments,
                    win->fragment_size,
                    win->shellout
                );
	    }
	}

        return(TRUE);
}
#undef QUIT_GTK_MAIN_RETURN

int main(int argc, char *argv[])
{
        gboolean initialized_gtk = FALSE;
	guint toid, toid2;
	GdkWindow *window;
	GtkWidget *toplevel, *w;
	connection_bar_struct *cb;
	win_struct *win;
	dspview_struct *dv;
	const gchar *y_server_address = "127.0.0.1:9433";


        /* Initialize GTK+ as needed. */
        if(!initialized_gtk)
        {
            if(!gtk_init_check(&argc, &argv))
            {
                fprintf(
                    stderr,
 "This program requires X.\n"
                );
                return(1);
            }
            initialized_gtk = TRUE;
        }

        gdk_rgb_init();

        imlib_handle = Imlib_init(GDK_DISPLAY());

	GUISetGlobalTipsState(TRUE);

        CDialogInit();
        CSDInit();


	/* Create window structure. */
	win = (win_struct *)g_malloc0(sizeof(win_struct));
	win->con = NULL;
	win->audio_mode_name = NULL;
	win->shellout = FALSE;

	/* Load background image. */
	win->bg_img = ImlibIOLoadXPMDataRGBA(
	    imlib_handle, (const gchar **)dspview_bg1_xpm,
	    &win->bg_img_width, &win->bg_img_height,
	    &win->bg_img_bpl
	);

	/* Create toplevel. */
	win->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_window_set_title(GTK_WINDOW(w), "YConsole");
/*	gtk_widget_set_usize(w, 320, -1); */
        gtk_window_set_policy(
            GTK_WINDOW(w),
            TRUE, TRUE, TRUE
        );
        gtk_widget_realize(w);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
	    GUISetWMIcon(window, (guint8 **)icon_yconsole_48x48_xpm);
        }
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(delete_cb), win
        );

	/* Create connection bar. */
	win->conbar = cb = ConBarNew(
	    CONBAR_DISPLAY_FLAG_WINDOW_MENU | CONBAR_DISPLAY_FLAG_LOCATION |
	    CONBAR_DISPLAY_FLAG_DSPVIEW | CONBAR_DISPLAY_FLAG_AUDIO |
	    CONBAR_DISPLAY_FLAG_REFRESH | CONBAR_DISPLAY_FLAG_SHELLOUT |
	    CONBAR_DISPLAY_FLAG_SHUTDOWN,
	    win,
	    connect_cb, yserver_list_cb, close_cb,
	    new_console_cb, new_audio_cb, new_mixer_cb, new_player_cb,
	    options_cb, refresh_cb, shellout_cb, shutdown_cb,
	    w
	);
	dv = ConBarGetDSPView(cb);
	ConBarMap(cb);

	DSPViewSetBG(
	    dv, win->bg_img,
	    win->bg_img_width, win->bg_img_height, win->bg_img_bpl, 4
	);
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_NOVA);
        DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
        DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);  /* 0xAABBGGRR. */

	/* Make initial connection to Y server and update values on
	 * all widgets.
	 */
	do_connect(win, y_server_address);

	/* Show toplevel window. */
	gtk_widget_show(toplevel);

        /* Set timeout callbacks. */
	win->toid = toid = gtk_timeout_add(
            30, (GtkFunction)timeout_cb, win
        );
        win->toid2 = toid2 = gtk_timeout_add(
            500, (GtkFunction)timeout2_cb, win
        );


	/* Enter main GTK+ loop. */
	gtk_main();

	/* Remove timeout callbacks. */
	if(win->toid != (guint)-1)
	{
	    gtk_timeout_remove(win->toid);
	    win->toid = (guint)-1;
	}
        if(win->toid2 != (guint)-1)
        {
            gtk_timeout_remove(win->toid2);
            win->toid2 = (guint)-1;
        }

	/* Destroy widgets. */
	do_disconnect(win, FALSE);
	ConBarDelete(win->conbar);
	gtk_widget_destroy(win->toplevel);
	g_free(win->bg_img);
	g_free(win);


        CDialogShutdown();
        CSDShutdown();

	return(0);
}
