#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "guiutils.h"
#include "cdialog.h"

#include "cfg.h"
#include "optwin.h"
#include "optwinop.h"
#include "yc.h"
#include "yccb.h"
#include "ycop.h"
#include "config.h"


static void OptWinFetchWidgetValueFromCFG(
	yc_core_struct *core_ptr, optwin_cfgref_struct *cfgref
);
static void OptWinSetWidgetValueToCFG(
	yc_core_struct *core_ptr, optwin_cfgref_struct *cfgref
);

void OptWinDoFetch(optwin_struct *optwin);
void OptWinDoApply(optwin_struct *optwin);



/*
 *	Nexus for fetching a value from the configruation list to the
 *	given widget.
 */
static void OptWinFetchWidgetValueFromCFG(
	yc_core_struct *core_ptr, optwin_cfgref_struct *cfgref
)
{
	gint i, cfg_type;
	const gchar *cfg_parm;
	cfg_item_struct *cfg_item;
	GtkWidget *w;

	if((core_ptr == NULL) || (cfgref == NULL))
	    return;

	cfg_parm = cfgref->cfg_parm;
	if(cfg_parm == NULL)
	    return;

	/* Match configuration item from the configuration list on
	 * the core
	 */
	i = CFGItemListMatchParameter(
	    core_ptr->cfg_list, cfg_parm
	);
	if(i < 0)
	    return;

	cfg_item = &core_ptr->cfg_list[i];
	cfg_type = cfg_item->type;


	/* Do special handling first */

	/* Stack list? */
	if(cfgref->slist != NULL)
	{
	    /* Configuration value type must be CFG_ITEM_TYPE_INTLIST
	     * for the stack list
	     */
	    if(cfg_type == CFG_ITEM_TYPE_INTLIST)
	    {
		cfg_intlist_struct *intlist = CFG_INTLIST(cfg_item->value);
		stack_list_struct *slist = cfgref->slist;

		StackListItemClearSrc(slist);
		StackListItemSetAllFromCacheSrc(slist);

		/* Remove any occurance of identical items in the
		 * intlist from the source list that do now allow
		 * multiple coppies
		 *
		 * This is so that no item already on the target
		 * list will appear on the source list
		 */
		if(intlist != NULL)
		{
		    GList *glist = intlist->list;
		    while(glist != NULL)
		    {
			StackListItemRemoveByIDSrc(
			    slist,
			    (gint)glist->data,
			    TRUE
			);
			glist = g_list_next(glist);
		    }
		}

		/* Add items to the stack list's target list from the
		 * intlist where each value is an id
		 */
		StackListItemClearTar(slist);
		if(intlist != NULL)
		{
		    GList *glist = intlist->list;
		    while(glist != NULL)
		    {
			StackListItemAppendTar(
			    slist, (gint)glist->data
			);
			glist = g_list_next(glist);
		    }
		}
	    }

	    return;
	}


	/* Standard widget handling, get widget that is to hold value */
	w = OPTWIN_CFGREF_WIDGET(cfgref);
	if(w == NULL)
	    return;

	/* Handle by widget type */
	if(GTK_IS_SPIN_BUTTON(w))
	{
	    GtkSpinButton *spin_button = GTK_SPIN_BUTTON(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = (const gchar *)cfg_item->value;
		if(values != NULL)
		    gtk_entry_set_text(GTK_ENTRY(w), values);
	    }
	    else
	    {
		gfloat valuef = CFGItemListGetValueF(
		    core_ptr->cfg_list, cfg_parm
		);
		gtk_spin_button_set_value(spin_button, valuef);
	    }
	}
	else if(GTK_IS_COMBO(w))
	{
	    GtkCombo *combo = GTK_COMBO(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = (const gchar *)cfg_item->value;
		if(values != NULL)
		    gtk_entry_set_text(GTK_ENTRY(combo->entry), values);
	    }
	    else
	    {
		/* All else assume handling as an enumeration, where
		 * the configuration value is an item index on the
		 * combo's list of strings
		 */
		const gchar *new_text;
		gint n = CFGItemListGetValueI(
		    core_ptr->cfg_list, cfg_parm
		);
		GList *glist = GUIComboGetList(w);

		/* Get glist pointer to the index specified by the
		 * cfg value
		 */
		glist = g_list_nth(glist, n);
		new_text = (const gchar *)((glist != NULL) ?
		    glist->data : NULL
		);
		if(new_text != NULL)
		    gtk_entry_set_text(GTK_ENTRY(combo->entry), new_text);
	    }
	}
	else if(GTK_IS_ENTRY(w))
	{
	    GtkEntry *entry = GTK_ENTRY(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = (const gchar *)cfg_item->value;
		if(values != NULL)
		    gtk_entry_set_text(entry, values);
	    }
	    else
	    {
		gchar num_str[80];

		*num_str = '\0';
		switch(cfg_type)
		{
		  case CFG_ITEM_TYPE_INT8:
		  case CFG_ITEM_TYPE_UINT8:
		  case CFG_ITEM_TYPE_INT16:
		  case CFG_ITEM_TYPE_UINT16:
		  case CFG_ITEM_TYPE_INT32:
		  case CFG_ITEM_TYPE_UINT32:
		    sprintf(
			num_str, "%i",
			CFGItemListGetValueI(
			    core_ptr->cfg_list, cfg_parm
			)
		    );
		    break;

		  case CFG_ITEM_TYPE_INT64:
		  case CFG_ITEM_TYPE_UINT64:
		    sprintf(
			num_str, "%ld",
			CFGItemListGetValueL(
			    core_ptr->cfg_list, cfg_parm
			)
		    );
		    break;

		  case CFG_ITEM_TYPE_FLOAT:
		    sprintf(
			num_str, "%f",
			CFGItemListGetValueF(
			    core_ptr->cfg_list, cfg_parm
			)
		    );
		    break;

		  case CFG_ITEM_TYPE_DOUBLE:
		    sprintf(
			num_str, "%f",
			CFGItemListGetValueD(
			    core_ptr->cfg_list, cfg_parm
			)
		    );
		    break;
		}
		gtk_entry_set_text(entry, num_str);
	    }
	}
	else if(GTK_IS_RADIO_BUTTON(w))
	{
	    gint valuei;
	    GtkToggleButton *toggle_button = GTK_TOGGLE_BUTTON(w);

	    valuei = CFGItemListGetValueI(core_ptr->cfg_list, cfg_parm);
	    if(valuei == cfgref->radio_value)
		toggle_button->active = TRUE;
	    else
		toggle_button->active = FALSE;
	}
	else if(GTK_IS_TOGGLE_BUTTON(w))
	{
	    gint valuei;
	    GtkToggleButton *toggle_button = GTK_TOGGLE_BUTTON(w);

	    valuei = CFGItemListGetValueI(core_ptr->cfg_list, cfg_parm);
	    toggle_button->active = valuei ? TRUE : FALSE;
	}
	else if(GTK_IS_RANGE(w))
	{
	    GtkRange *range = GTK_RANGE(w);
	    GtkAdjustment *adj = range->adjustment;
	    if(adj != NULL)
		adj->value = CFGItemListGetValueF(
		    core_ptr->cfg_list, cfg_parm
		);
	}
	else if(GTK_IS_DRAWING_AREA(w))
	{
	    const cfg_color_struct *cs = CFGItemListGetValueColor(
		core_ptr->cfg_list, cfg_parm
	    );
	    if(cs != NULL)
	    {
		GdkColor ct;
		GdkWindow *window;

		ct.red		= (gushort)-1 * cs->r;
		ct.green	= (gushort)-1 * cs->g;
		ct.blue		= (gushort)-1 * cs->b;

		OptWinCfgRefSetColor(cfgref, &ct, &ct);
		window = w->window;
		if(window != NULL)
		    gdk_window_set_background(window, &cfgref->fg_color);
	    }
	}
	else
	{
		/* Unsupported widget*/
	}
}

/*
 *      Nexus for setting a value from the given widget to the configuration
 *	list.
 *
 *      Inputs assumed valid.
 */
static void OptWinSetWidgetValueToCFG(
	yc_core_struct *core_ptr, optwin_cfgref_struct *cfgref
)
{
	gint i, cfg_type;
	const gchar *cfg_parm;
	cfg_item_struct *cfg_item;
	GtkWidget *w;


	if((core_ptr == NULL) || (cfgref == NULL))
	    return;

	cfg_parm = cfgref->cfg_parm;
	if(cfg_parm == NULL)
	    return;

	/* Match configuration item from the configuration list on
	 * the core
	 */
	i = CFGItemListMatchParameter(
	    core_ptr->cfg_list, cfg_parm
	);
	if(i < 0)
	    return;

	cfg_item = &core_ptr->cfg_list[i];
	cfg_type = cfg_item->type;


	/* Do special handling first */

	/* Stack list? */
	if(cfgref->slist != NULL)
	{
	    /* Configuration value type must be CFG_ITEM_TYPE_INTLIST
	     * for the stack list.
	     */
	    if(cfg_type == CFG_ITEM_TYPE_INTLIST)
	    {
		gint *new_id;
		gint total_new_ids;
		cfg_intlist_struct *intlist =
		    (cfg_intlist_struct *)cfg_item->value;
		stack_list_struct *slist = cfgref->slist;

		/* Get listing of ids in the stack list's target list and
		 * set those ids to the intlist
		 */
		new_id = StackListItemGetTar(slist, &total_new_ids);
		if((new_id != NULL) && (intlist != NULL))
		{
		    gint i;

		    g_list_free(intlist->list);
		    intlist->list = NULL;

		    /* Set the new list of stack item IDs to the
		     * intlist
		     */
		    for(i = 0; i < total_new_ids; i++)
			intlist->list = g_list_append(
			    intlist->list,
			    (gpointer)new_id[i]
			);
		}

		g_free(new_id);
	    }
	}


	/* Standard widget handling, get widget that is to hold value */
	w = OPTWIN_CFGREF_WIDGET(cfgref);
	if(w == NULL)
	    return;

	/* Handle by widget type */
	if(GTK_IS_SPIN_BUTTON(w))
	{
	    GtkSpinButton *spin_button = GTK_SPIN_BUTTON(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = gtk_entry_get_text(GTK_ENTRY(w));
		if(values != NULL)
		    CFGItemListSetValueS(
			core_ptr->cfg_list, cfg_parm,
			values, FALSE
		    );
	    }
	    else
	    {
		if(spin_button->digits > 0)
		{
		    gfloat valuef = gtk_spin_button_get_value_as_float(
			spin_button
		    );
		    CFGItemListSetValueF(
			core_ptr->cfg_list, cfg_parm,
			valuef, FALSE
		    );
		}
		else
		{
		    gfloat valuei = gtk_spin_button_get_value_as_int(
			spin_button
		    );
		    CFGItemListSetValueI(
			core_ptr->cfg_list, cfg_parm,
			valuei, FALSE
		    );
		}
	    }
	}
	else if(GTK_IS_COMBO(w))
	{
	    GtkCombo *combo = GTK_COMBO(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = gtk_entry_get_text(GTK_ENTRY(combo->entry));
		if(values != NULL)
		    CFGItemListSetValueS(
			core_ptr->cfg_list, cfg_parm,
			values, FALSE
		    );
	    }
	    else
	    {
		/* All else assume handling as an enumeration, where
		 * the current string value should match a string in the
		 * combo's list of strings. The matched string index
		 * value will be used as the new configuration value.
		 */
		const gchar *sel_text = gtk_entry_get_text(
		    GTK_ENTRY(combo->entry)
		);
		gint string_index = 0;
		GList *glist = GUIComboGetList(w);


		while(glist != NULL)
		{
		    if((glist->data != NULL) && (sel_text != NULL))
		    {
			if(!strcasecmp((const gchar *)glist->data, sel_text))
			    break;
		    }
		    glist = glist->next;
		    string_index++;
		}

		CFGItemListSetValueI(
		    core_ptr->cfg_list, cfg_parm, string_index, FALSE
		);
	    }
	}
	else if(GTK_IS_ENTRY(w))
	{
	    GtkEntry *entry = GTK_ENTRY(w);

	    if(cfg_type == CFG_ITEM_TYPE_STRING)
	    {
		const gchar *values = gtk_entry_get_text(entry);
		if(values != NULL)
		    CFGItemListSetValueS(
			core_ptr->cfg_list, cfg_parm,
			values, FALSE
		    );
	    }
	    else
	    {
		const gchar *cstrptr;
		const gchar *values = gtk_entry_get_text(entry);
		if(values != NULL)
		{
		    /* Decimal in the string? */
		    cstrptr = strchr(values, '.');
		    if(cstrptr != NULL)
			CFGItemListSetValueF(
			    core_ptr->cfg_list, cfg_parm,
			    atof(values), FALSE
			);
		    else
			CFGItemListSetValueI(
			    core_ptr->cfg_list, cfg_parm,
			    atoi(values), FALSE
			);
		}
	    }
	}
	else if(GTK_IS_RADIO_BUTTON(w))
	{
	    GtkToggleButton *toggle_button = GTK_TOGGLE_BUTTON(w);

	    if(toggle_button->active)
		CFGItemListSetValueI(
		    core_ptr->cfg_list, cfg_parm,
		    cfgref->radio_value, FALSE
		);
	}
	else if(GTK_IS_TOGGLE_BUTTON(w))
	{
	    GtkToggleButton *toggle_button = GTK_TOGGLE_BUTTON(w);

	    CFGItemListSetValueI(
		core_ptr->cfg_list, cfg_parm,
		toggle_button->active ? TRUE : FALSE, FALSE
	    );
	}
	else if(GTK_IS_RANGE(w))
	{
	    GtkRange *range = GTK_RANGE(w);
	    GtkAdjustment *adj = range->adjustment;
	    switch(cfg_type)
	    {
	      case CFG_ITEM_TYPE_FLOAT:
	      case CFG_ITEM_TYPE_DOUBLE:
		CFGItemListSetValueF(
		    core_ptr->cfg_list, cfg_parm,
		    adj->value, FALSE
		);
		break;
	      default:
		CFGItemListSetValueI(
		    core_ptr->cfg_list, cfg_parm,
		    (gint)adj->value, FALSE
		);
		break;
	    }
	}
	else if(GTK_IS_DRAWING_AREA(w))
	{
	    const GdkColor *cs = &cfgref->fg_color;
	    cfg_color_struct ct;

	    ct.r = (gfloat)cs->red	/ (gfloat)((gushort)-1);
	    ct.g = (gfloat)cs->green	/ (gfloat)((gushort)-1);
	    ct.b = (gfloat)cs->blue	/ (gfloat)((gushort)-1);
	    ct.a = 1.0f;

	    if(cfg_type == CFG_ITEM_TYPE_COLOR)
		CFGItemListSetValueColor(
		    core_ptr->cfg_list, cfg_parm,
		    &ct, FALSE
		);
	}
	else
	{
	    /* Unsupported widget*/
	}
}

/*
 *	Fetches values from configuration list filtered by the cfgrefs on
 *	the given optwin and placed into the widgets.
 */
void OptWinDoFetch(optwin_struct *optwin)
{
	gint i;
	yc_core_struct *core_ptr;


	if(optwin == NULL)
	    return;

	core_ptr = YC_CORE(optwin->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Iterate through all cfgrefs*/
	for(i = 0; i < optwin->total_cfgrefs; i++)
	    OptWinFetchWidgetValueFromCFG(core_ptr, optwin->cfgref[i]);
}

/*
 *	Applies values from the widgets specified by the cfgrefs on the
 *	given optwin into the configuration list.
 */
void OptWinDoApply(optwin_struct *optwin)
{
	gint i;
	yc_core_struct *core_ptr;


	if(optwin == NULL)
	    return;

	core_ptr = YC_CORE(optwin->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Iterate through all cfgrefs*/
	for(i = 0; i < optwin->total_cfgrefs; i++)
	    OptWinSetWidgetValueToCFG(
		core_ptr, optwin->cfgref[i]
	    );

	/* Send out reconfigured signal to all resources*/
	YCReconfiguredEmit(core_ptr);
}
