#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "startupdialog.h"
#include "cfg.h"
#include "yc.h"
#include "yccfglist.h"
#include "config.h"


#include "images/icon_y_48x48.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/start_yserver_banner_800x48.xpm"


static void StartupDialogDeleteCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void StartupDialogCancelCB(GtkWidget *widget, gpointer data);
static void StartupDialogCloseCB(GtkWidget *widget, gpointer data);
static void StartupDialogTerminationCB(
	GtkWidget *widget, gint pid, gpointer data
);

gint StartupDialogExecute(
	startup_dialog_struct *sd, const gchar *cmd
);
void StartupDialogSavePositions(startup_dialog_struct *sd);

startup_dialog_struct *StartupDialogNew(
	const gchar *title, guint8 **wm_icon_xpm_data,
	guint8 **heading_xpm_data,
	gpointer core_ptr
);
void StartupDialogMap(startup_dialog_struct *sd);
void StartupDialogUnmap(startup_dialog_struct *sd);
void StartupDialogDelete(startup_dialog_struct *sd);


#define BTN_WIDTH       (100 + (2 * 3))
#define BTN_HEIGHT      (30 + (2 * 3))



/*
 *	Startup dialog toplevel "delete_event" signal callback.
 */
static void StartupDialogDeleteCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	StartupDialogCancelCB(widget, data);
}

/*
 *      Startup dialog cancel callback.
 */
static void StartupDialogCancelCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	startup_dialog_struct *sd = STARTUP_DIALOG(data);
	if(sd == NULL)
	    return;

	w = STARTUP_DIALOG_PROCCON(sd);
	ProcConKill(w);
	StartupDialogSavePositions(sd);
	StartupDialogUnmap(sd);
}

/*
 *      Startup dialog close callback.
 */
static void StartupDialogCloseCB(GtkWidget *widget, gpointer data)
{
	startup_dialog_struct *sd = STARTUP_DIALOG(data);
	if(sd == NULL)
	    return;

	StartupDialogSavePositions(sd);
	StartupDialogUnmap(sd);
}

/*
 *	Process termination callback.
 */
static void StartupDialogTerminationCB(
	GtkWidget *widget, gint pid, gpointer data
)
{
	GtkWidget *w;
	startup_dialog_struct *sd = STARTUP_DIALOG(data);
	if(sd == NULL)
	    return;

	w = sd->close_on_termination_tb;
	if(GTK_TOGGLE_BUTTON(w)->active)
	{
	    /* Automatically close on termination */
	    StartupDialogSavePositions(sd);
	    StartupDialogUnmap(sd);
	}
	else
	{
	    /* Do not automatically close on termination, instead remap
	     * cancel and close buttons.
	     */
	    w = sd->cancel_btn;
	    if(w != NULL)
		gtk_widget_hide(w);
	    w = sd->close_btn;
	    if(w != NULL)
		gtk_widget_show(w);
	}
}


/*
 *	Executes the given command.
 *
 *	Returns the pid of the process or 0 on error.
 */
gint StartupDialogExecute(
	startup_dialog_struct *sd, const gchar *cmd
)
{
	gint pid = 0;
	GtkWidget *w;


	if((sd == NULL) || (cmd == NULL))
	    return(pid);

	w = STARTUP_DIALOG_PROCCON(sd);
	if((w == NULL) || (*cmd == '\0'))
	    return(pid);

	pid = ProcConExecuteNotify(
	    w, cmd, sd, StartupDialogTerminationCB
	);

	return(pid);
}

/*
 *      Save current position value of window to configuration list on
 *      core structure.
 */
void StartupDialogSavePositions(startup_dialog_struct *sd)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(sd == NULL)
	    return;

	core_ptr = YC_CORE(sd->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_STARTUP_DIALOG_SAVE_POSITIONS
	))
	    return;

	/* Save toplevel position */
	w = STARTUP_DIALOG_TOPLEVEL(sd);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_STARTUP_DIALOG_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_STARTUP_DIALOG_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_STARTUP_DIALOG_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_STARTUP_DIALOG_HEIGHT, height);
	}

	/* Close on termination toggle button */
	w = sd->close_on_termination_tb;
	if(w != NULL)
	    YC_SET_CFG_I(
		core_ptr, YC_CFG_PARM_STARTUP_DIALOG_AUTO_CLOSE,
		GTK_TOGGLE_BUTTON(w)->active
	    );

}


/*
 *	Creates a new startup dialog.
 */
startup_dialog_struct *StartupDialogNew(
	const gchar *title, guint8 **wm_icon_xpm_data,
	guint8 **heading_xpm_data,
	gpointer core_ptr
)
{
	const gint	border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkStyle *style;
	GtkWidget *w, *parent, *parent2, *toplevel;
	yc_core_struct *cp = YC_CORE(core_ptr);
	startup_dialog_struct *sd = STARTUP_DIALOG(g_malloc0(
	    sizeof(startup_dialog_struct)
	));
	if(sd == NULL)
	    return(NULL);

	/* Reset values */
	sd->initialized = TRUE;
	sd->map_state = FALSE;
	sd->accelgrp = accelgrp = gtk_accel_group_new();
	sd->core_ptr = core_ptr;

	/* Create toplevel */
	sd->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_name(w, "startup dialog");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_STARTUP_DIALOG_SAVE_POSITIONS))
	    {
		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_STARTUP_DIALOG_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_STARTUP_DIALOG_Y)
		);
		gtk_widget_set_usize(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_STARTUP_DIALOG_WIDTH),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_STARTUP_DIALOG_HEIGHT)
		);
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_window_set_title(
	    GTK_WINDOW(w), (title != NULL) ? title : "Start Up"
	);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		GDK_DECOR_TITLE |  GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_RESIZE |
		GDK_FUNC_MINIMIZE | GDK_FUNC_MAXIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(
		window, (wm_icon_xpm_data != NULL) ?
		    wm_icon_xpm_data : (guint8 **)icon_y_48x48_xpm
	    );
	}
	style = gtk_widget_get_style(w);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(StartupDialogDeleteCB), sd
	);
	parent = w;

	/* Main vbox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Frame for heading */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Drawing area for heading */
	w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, -1, 48);
	gtk_widget_add_events(
	    w,
	    GDK_EXPOSURE_MASK | GDK_STRUCTURE_MASK
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_realize(w);
	gtk_widget_show(w);
	/* Heading pixmap */
	if((style != NULL) && (w->window != NULL))
	{
	    GdkBitmap *mask;
	    GdkPixmap *pixmap = gdk_pixmap_create_from_xpm_d(
		w->window, &mask,
		&style->bg[GTK_STATE_NORMAL],
		(gchar **)start_yserver_banner_800x48_xpm
	    );
	    if(pixmap != NULL)
	    {
		gdk_window_set_back_pixmap(w->window, pixmap, FALSE);
		gdk_pixmap_unref(pixmap);
		if(mask != NULL)
		    gdk_bitmap_unref(mask);
	    }
	}

	/* Process console */
	sd->proccon = w = ProcConNew(
	    PROCCON_DISPLAY_STDOUT | PROCCON_DISPLAY_STDERR
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);


	/* Hbox for buttons */
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Close on termination toggle button */
	sd->close_on_termination_tb = w =
	    gtk_check_button_new_with_label("Close When Finished");
	GTK_TOGGLE_BUTTON(w)->active = YC_GET_CFG_I(
	    cp, YC_CFG_PARM_STARTUP_DIALOG_AUTO_CLOSE
	) ? TRUE : FALSE;
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Cancel button */
	sd->cancel_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm, "Cancel", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(w, BTN_WIDTH, BTN_HEIGHT);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(StartupDialogCancelCB), sd
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);

	/* Close button */
	sd->close_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_close_20x20_xpm, "Close", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(w, BTN_WIDTH, BTN_HEIGHT);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(StartupDialogCloseCB), sd
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	

	return(sd);
}

/*
 *	Maps the startup dialog and resets the map states for the
 *	cancel and close buttons.
 */
void StartupDialogMap(startup_dialog_struct *sd)
{
	if(sd == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = STARTUP_DIALOG_TOPLEVEL(sd);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    sd->map_state = TRUE;

	    /* Reset map states for cancel and close buttons */
	    w = sd->cancel_btn;
	    if(w != NULL)
		gtk_widget_show(w);
	    w = sd->close_btn;
	    if(w != NULL)
		gtk_widget_hide(w);
	}
}

/*
 *	Unmaps the startup dialog.
 */
void StartupDialogUnmap(startup_dialog_struct *sd)
{
	if(sd == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = STARTUP_DIALOG_TOPLEVEL(sd);
	    if(w != NULL)
		gtk_widget_hide(w);
	    sd->map_state = FALSE;
	}
}

/*
 *	Deletes the startup dialog.
 */
void StartupDialogDelete(startup_dialog_struct *sd)
{
	GtkWidget **w;


	if(sd == NULL)
	    return;

	if(sd->initialized)
	{
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) \
   { gtk_widget_destroy(*w); *w = NULL; } \
}

	    /* Destroy widgets */
	    w = &sd->close_btn;
	    DO_DESTROY_WIDGET
	    w = &sd->cancel_btn;
	    DO_DESTROY_WIDGET
	    w = &sd->close_on_termination_tb;
	    DO_DESTROY_WIDGET
	    w = &sd->proccon;
	    DO_DESTROY_WIDGET
	    w = &sd->heading_pm;
	    DO_DESTROY_WIDGET
	    w = &sd->toplevel;
	    DO_DESTROY_WIDGET

	    if(sd->accelgrp != NULL)
	    {
		gtk_accel_group_unref(sd->accelgrp);
		sd->accelgrp = NULL;
	    }

#undef DO_DESTROY_WIDGET
	}

	/* Deallocate structure itself */
	g_free(sd);
}
