#include <Y2/Y.h>
#include <Y2/Ylib.h>
#include <gtk/gtk.h>
#include "ycanalyze.h"


void YCAnalyzeAverageChannelVolume(
	const guint8 *buf, gint buf_len,
	gfloat *channel, gint channels
);
const YAudioModeValuesStruct *YCMatchAudioModeByValues(
	YAudioModeValuesStruct **list, int total,
	int sample_rate, int sample_size, int channels
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)             (((x) < 0) ? ((x) * -1) : (x))


/*
 *	Scans the given buffer and calculates the volume of the buffer
 *	as a coefficient from 0.0 to 1.0 for each channel.  The number
 *	of channels can be 1, 2, or 4.
 */
void YCAnalyzeAverageChannelVolume(
	const guint8 *buf, gint buf_len,
	gfloat *channel, gint channels
)
{
	if((buf == NULL) || (buf_len <= 0))
	    return;

	if(channels == 4)
	{
	    const guint8 *buf_ptr, *buf_end;
	    gint cbuf_len = buf_len / channels;
	    gulong v_avg[4];

	    v_avg[0] = 0;
	    v_avg[1] = 0;
	    v_avg[2] = 0;
	    v_avg[3] = 0;

	    /* Scan buffer and get v_avg values from 0x00 to 0x80. */
	    for(buf_ptr = buf,
		buf_end = buf_ptr + buf_len;
		buf_ptr < buf_end;
		buf_ptr++
	    )
	    {
		/* Left channel. */
		v_avg[0] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);

		/* Right channel. */
		buf_ptr++;
		if(buf_ptr < buf_end)
		    v_avg[1] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);

		/* Back left channel. */
		buf_ptr++;
		if(buf_ptr < buf_end)
		    v_avg[2] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);

		/* Back right channel. */
		buf_ptr++;
		if(buf_ptr < buf_end)
		    v_avg[3] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);
	    }

	    /* Calculate channel volume coefficients. */
	    channel[0] = MIN(
		(gfloat)(v_avg[0] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	    channel[1] = MIN(
		(gfloat)(v_avg[1] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	    channel[2] = MIN(
		(gfloat)(v_avg[2] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	    channel[3] = MIN(
		(gfloat)(v_avg[3] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	}
	else if(channels == 2)
	{
	    const guint8 *buf_ptr, *buf_end;
	    gint cbuf_len = buf_len / channels;
	    gulong v_avg[2];

	    v_avg[0] = 0;
	    v_avg[1] = 0;

	    /* Scan buffer and get v_avg values from 0x00 to 0x80. */
	    for(buf_ptr = buf,
		buf_end = buf_ptr + buf_len;
		buf_ptr < buf_end;
		buf_ptr++
	    )
	    {
		/* Left channel. */
		v_avg[0] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);

		/* Right channel. */
		buf_ptr++;
		if(buf_ptr < buf_end)
		    v_avg[1] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);
	    }

	    /* Calculate channel volume coefficients. */
	    channel[0] = MIN(
		(gfloat)(v_avg[0] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	    channel[1] = MIN(
		(gfloat)(v_avg[1] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	}
	else if(channels == 1)
	{
	    const guint8 *buf_ptr, *buf_end;
	    gint cbuf_len = buf_len / channels;
	    gulong v_avg[1];

	    v_avg[0] = 0;

	    /* Scan buffer and get v_avg values from 0x00 to 0x80. */
	    for(buf_ptr = buf,
		buf_end = buf_ptr + buf_len;
		buf_ptr < buf_end;
		buf_ptr++
	    )
		v_avg[0] += (gulong)ABSOLUTE((gint32)*buf_ptr - 0x80);

	    /* Calculate channel volume coefficients. */
	    channel[0] = MIN(
		(gfloat)(v_avg[0] / (gulong)cbuf_len) / (gfloat)0x80,
		1.0f
	    );
	}
}

/*
 *	Returns a pointer to the Audio preset mode that matches
 *	the given values from the list of Audio preset modes.  The
 *	returned pointer will point to an Audio preset mode in the
 *	given list or NULL if there is no match.
 */
const YAudioModeValuesStruct *YCMatchAudioModeByValues(
	YAudioModeValuesStruct **list, int total,
	int sample_rate, int sample_size, int channels
)
{
	int i;
	const YAudioModeValuesStruct *m;

	for(i = 0; i < total; i++)
	{
	    m = list[i];
	    if(m == NULL)
		continue;

	    if((sample_rate > 0) &&
	       (sample_rate != m->sample_rate)
	    )
		continue;
	    if((sample_size > 0) &&
	       (sample_size != m->sample_size)
	    )
		continue;
	    if((channels > 0) &&
	       (channels != m->channels)
	    )
		continue;

	    return(m);
	}

	return(NULL);
}
