#include <gtk/gtk.h>

#include "guiutils.h"
#include "guirgbimg.h"

#include "../include/imlibio.h"

#include "yc.h"
#include "ycimg.h"

#include "images/dspview_bg_default.xpm"
#include "images/dspview_bg_default_small.xpm"


void YCImgLoadDefault(yc_core_struct *core_ptr);
void YCImgLoad(
	yc_core_struct *core_ptr,
	const gchar *large_img_path,
	gint large_img_fit_method,
	const gchar *small_img_path,
	gint small_img_fit_method
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)             (((x) < 0) ? ((x) * -1) : (x))


/*
 *	Loads the default DSPView image on to the core structure.
 *
 *	The image buffers on the core structure will be reallocated.
 */
void YCImgLoadDefault(yc_core_struct *core_ptr)
{
	if(core_ptr == NULL)
	    return;

	if(core_ptr->imlib_handle == NULL)
	    return;

	/* Reallocate large DSPView image buffer and load default
	 * image.
	 */
	g_free(core_ptr->dsp_bg_img);
	core_ptr->dsp_bg_img = ImlibIOLoadXPMDataRGBA(
	    core_ptr->imlib_handle, (const gchar **)dspview_bg_default_xpm,
	    &core_ptr->dsp_bg_img_width, &core_ptr->dsp_bg_img_height,
	    &core_ptr->dsp_bg_img_bpl
	);

	/* Reallocate small DSPView image buffer and load default
	 * image.
	 */
	g_free(core_ptr->dsp_bg_img_small);
	core_ptr->dsp_bg_img_small = ImlibIOLoadXPMDataRGBA(
	    core_ptr->imlib_handle, (const gchar **)dspview_bg_default_small_xpm,
	    &core_ptr->dsp_bg_img_small_width,
	    &core_ptr->dsp_bg_img_small_height,
	    &core_ptr->dsp_bg_img_small_bpl
	);
}

/*
 *	Loads new DSPView images on to the core structure.
 *
 *	The DSPView background image buffers on the core structure must
 *	already be allocated.  They will not be reallocated by this
 *	call.
 */
void YCImgLoad(
	yc_core_struct *core_ptr,
	const gchar *large_img_path,
	gint large_img_fit_method,
	const gchar *small_img_path,
	gint small_img_fit_method
)
{
	if(core_ptr == NULL)
	    return;

	if(core_ptr->imlib_handle == NULL)
	    return;

#define DO_LOAD_IMAGE						\
{ 								\
	    /* Load source image. */				\
	    if((path != NULL) ? (*path != '\0') : FALSE)	\
		sbuf = ImlibIOLoadImageFileRGBA(		\
		    core_ptr->imlib_handle, path,		\
		    &swidth, &sheight, &sbpl			\
		);						\
	    else						\
		sbuf = ImlibIOLoadXPMDataRGBA(			\
		    core_ptr->imlib_handle,			\
		    xpm_data,					\
		    &swidth, &sheight, &sbpl			\
		);						\
	    /* Source image loaded successfully? */		\
	    if((sbuf != NULL) &&				\
	       (swidth > 0) && (sheight > 0) && (sbpl > 0)	\
	    )							\
	    {							\
		/* How to copy source image to target image. */	\
		switch(fit_method)				\
		{						\
		  case YC_IMG_FIT_METHOD_SCALE:			\
		    GUIImageBufferResize(			\
			4,					\
			(const gpointer)sbuf,			\
			swidth, sheight, sbpl,			\
			tbuf, twidth, theight, tbpl		\
		    );						\
		    break;					\
		  case YC_IMG_FIT_METHOD_TILE:			\
		    for(y = 0; y < theight; y++)		\
		    {						\
			for(x = 0; x < twidth; x++)		\
			{					\
			    sbuf_ptr32 = (guint32 *)&sbuf[	\
				((y % sheight) * sbpl) +	\
				((x % swidth) * 4)		\
			    ];					\
			    tbuf_ptr32 = (guint32 *)&tbuf[	\
				(y * tbpl) + (x * 4)		\
			    ];					\
			    *tbuf_ptr32 = *sbuf_ptr32;		\
			}					\
		    }						\
		    break;					\
		}						\
	    }							\
	    else						\
	    {							\
		/* Source image not loaded, so just clear the	\
		 * target image.				\
		 */						\
		for(y = 0; y < theight; y++)			\
		{						\
		    for(x = 0; x < twidth; x++)			\
		    {						\
			tbuf_ptr32 = (guint32 *)&tbuf[		\
			    (y * tbpl) + (x * 4)		\
			];					\
			*tbuf_ptr32 = 0xff000000;		\
		    }						\
		}						\
	    }							\
	    /* Deallocate source image. */			\
	    g_free(sbuf);					\
	    sbuf = NULL;					\
	    swidth = sheight = sbpl = 0;			\
}


	/* Load large image. */
	if(core_ptr->dsp_bg_img != NULL)
	{
	    gint fit_method = large_img_fit_method;
	    const gchar *path = large_img_path;
	    gint x, y;
	    guint8 *tbuf = core_ptr->dsp_bg_img;
	    guint32 *tbuf_ptr32;
	    gint        twidth = core_ptr->dsp_bg_img_width,
			theight = core_ptr->dsp_bg_img_height,
			tbpl = core_ptr->dsp_bg_img_bpl;
	    guint8 *sbuf = NULL;
	    guint32 *sbuf_ptr32;
	    gint swidth = 0, sheight = 0, sbpl = 0;
	    const gchar **xpm_data =
		(const gchar **)dspview_bg_default_xpm;

	    DO_LOAD_IMAGE
	}

	/* Load large image. */
	if(core_ptr->dsp_bg_img_small != NULL)
	{
	    gint fit_method = small_img_fit_method;
	    const gchar *path = small_img_path;
	    gint x, y;
	    guint8 *tbuf = core_ptr->dsp_bg_img_small;
	    guint32 *tbuf_ptr32;
	    gint        twidth = core_ptr->dsp_bg_img_small_width,
			theight = core_ptr->dsp_bg_img_small_height,
			tbpl = core_ptr->dsp_bg_img_small_bpl;
	    guint8 *sbuf = NULL;
	    guint32 *sbuf_ptr32;
	    gint swidth = 0, sheight = 0, sbpl = 0;
	    const gchar **xpm_data =
		(const gchar **)dspview_bg_default_small_xpm;

	    DO_LOAD_IMAGE
	}


#undef DO_LOAD_IMAGE
}
