#include <stdlib.h>
#include <string.h>

#include <Y2/Y.h>
#include <Y2/Ylib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "guiutils.h"

#include "cfg.h"
#include "conbar.h"
#include "yconsolewin.h"
#include "yconsolecb.h"
#include "yc.h"
#include "ycop.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_yconsole_48x48.xpm"


void YConsoleSetTitle(yconsole_struct *yc, const gchar *s);
gchar *YConsoleGetLocation(yconsole_struct *yc);
gint YConsoleDoConnect(
	yconsole_struct *yc,
	const gchar *address, const gchar *start_command
);
void YConsoleDoDisconnect(yconsole_struct *yc, gboolean force_shutdown);
void YConsoleSavePosition(yconsole_struct *yc);

yconsole_struct *YConsoleNew(gpointer core_ptr);
void YConsoleUpdateMenus(yconsole_struct *yc);
void YConsoleSetBusy(yconsole_struct *yc, gboolean is_busy);
void YConsoleMap(yconsole_struct *yc);
void YConsoleUnmap(yconsole_struct *yc);
void YConsoleDelete(yconsole_struct *yc);


/* 
 *	Sets the title string to s, if s is NULL then values from the
 *	yconsole will be used to form the title.
 */
void YConsoleSetTitle(yconsole_struct *yc, const gchar *s)
{
	gchar *title;
	GtkWidget *w;


	if(yc == NULL)
	    return;

	w = YCONSOLE_TOPLEVEL(yc);
	if(w == NULL)
	    return;

	/* Title string given? */
	if(s != NULL)
	{
	    gtk_window_set_title(GTK_WINDOW(w), s);
	    return;
	}

	/* Format title string. */
	if(YCONSOLE_IS_CONNECTED(yc))
	{
	    s = ConBarGetLocation(YCONSOLE_CONBAR(yc));
	    title = g_strdup_printf("YConsole: %s", s);
	}
	else
	{
	    title = g_strdup("YConsole");
	}
	gtk_window_set_title(GTK_WINDOW(w), title);

	g_free(title);
}


/*
 *	Returns the current location.
 */
gchar *YConsoleGetLocation(yconsole_struct *yc)
{
	return((yc != NULL) ? ConBarGetLocation(YCONSOLE_CONBAR(yc)) : NULL);
}

/*
 *	Connects the yconsole to the Y server at the addres s.  If the
 *	yconsole is already connected then it will be disconneted first.
 *
 *	Returns non-zero on error and 0 on success.
 */
gint YConsoleDoConnect(
	yconsole_struct *yc,
	const gchar *address, const gchar *start_command
)
{
	YConnection *con;


	if(yc == NULL)
	    return(-1);

	/* Disconnect from Y server as needed. */
	YConsoleDoDisconnect(yc, FALSE);

	/* If no address is given then only perform disconnect. */
	if(address == NULL)
	    return(0);

	/* Connect to Y server. */
	yc->con = con = YOpenConnection(NULL, address);
	/* If not able to connect on first try and a server start
	 * command is given, then execute that command and try
	 * connecting again.
	 */
	if((con == NULL) && (start_command != NULL))
	{
	    /* Start server (with dialog) and try connecting again. */
	    YCNewStartupDialog(
		YC_CORE(yc->core_ptr), YCONSOLE_TOPLEVEL(yc),
		start_command, TRUE
	    );
	    yc->con = con = YOpenConnection(NULL, address);
	}
	if(con != NULL)
	{
	    /* Connected to Y server, now get some initial values. */
	    YEventAudioStats stat_buf;
	    dspview_struct *dv = ConBarGetDSPView(YCONSOLE_CONBAR(yc));

	    /* Update connection bar connected state. */
	    ConBarSetConnectState(YCONSOLE_CONBAR(yc), TRUE);

	    /* YSHM sound buffer pointer. */
	    yc->yshm_buf = (guint8 *)YSHMOpenSound(
		con, &yc->yshm_buf_len, &yc->yshm_id
	    );

	    /* Audio stats. */
	    if(!YGetAudioStats(con, &stat_buf))
	    {
		g_free(yc->audio_mode_name);
		yc->audio_mode_name = g_strdup(stat_buf.mode_name);
		yc->sample_rate = stat_buf.sample_rate;
		yc->sample_size = stat_buf.sample_size;
		yc->channels = stat_buf.channels;
		yc->cycle_us = stat_buf.cycle_us;
		yc->cycle_load = 0.0;
		yc->num_fragments = stat_buf.num_fragments;
		yc->fragment_size = stat_buf.fragment_size;
		yc->shellout = FALSE;

		/* Update audio values on connection bar. */
		YCONSOLE_SET_CONBAR_AUDIO(yc);

		ConBarSetShellOut(YCONSOLE_CONBAR(yc), yc->shellout);
		DSPViewSetChannels(dv, yc->channels);
	    }

	    /* Record location history on connection bar. */
	    ConBarSetLocation(YCONSOLE_CONBAR(yc), address, FALSE, TRUE);

	    /* Update title. */
	    YConsoleSetTitle(yc, NULL);
	}

	return(0);
}

/*
 *	Disconnects the yconsole from the Y server.
 */
void YConsoleDoDisconnect(yconsole_struct *yc, gboolean force_shutdown)
{
	dspview_struct *dv;


	if(yc == NULL)
	    return;


	/* Update connection bar connected state. */
	ConBarSetConnectState(YCONSOLE_CONBAR(yc), FALSE);


	/* Close YSHM sound buffer. */
	YSHMCloseSound(YCONSOLE_CONNECTION(yc), yc->yshm_buf);
	yc->yshm_buf = NULL;
	yc->yshm_buf_len = 0;
	yc->yshm_id = -1;

	/* Close connection or shutdown? */
	if(force_shutdown)
	    YShutdownServer(YCONSOLE_CONNECTION(yc));
	else
	    YCloseConnection(YCONSOLE_CONNECTION(yc), False);
	yc->con = NULL;


	/* Reset Y audio values. */
	memset(yc->channel_volume_avg, 0x00, 4 * sizeof(gfloat));
	yc->start_time = 0;
	g_free(yc->audio_mode_name);
	yc->audio_mode_name = NULL;
	yc->sample_rate = 0;
	yc->sample_size = 0;
	yc->channels = 0;
	yc->cycle_us = 0;
	yc->cycle_load = 0.0;
	yc->num_fragments = 0;
	yc->fragment_size = 0;
	yc->flip_stereo = FALSE;
	yc->shellout = FALSE;
	yc->version_major = yc->version_minor = 0;
	g_free(yc->server_name);
	yc->server_name = NULL;
	g_free(yc->sound_name);
	yc->sound_name = NULL;

	/* Update audio values on connection bar. */
	YCONSOLE_SET_CONBAR_AUDIO(yc);

	dv = ConBarGetDSPView(YCONSOLE_CONBAR(yc));
	DSPViewSetChannels(dv, yc->channels);

	/* Update title. */
	YConsoleSetTitle(yc, NULL);
}


/*
 *	Save current position value of window to configuration list on
 *	core structure.
 */
void YConsoleSavePosition(yconsole_struct *yc)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(yc == NULL)
	    return;

	core_ptr = YC_CORE(yc->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_CONSOLE_SAVE_POSITIONS
	))
	    return;

	/* Save toplevel position. */
	w = YCONSOLE_TOPLEVEL(yc);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_CONSOLE_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_CONSOLE_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_CONSOLE_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_CONSOLE_HEIGHT, height);
	}
}


/*
 *	Create a new yconsole.
 */
yconsole_struct *YConsoleNew(gpointer core_ptr)
{
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *toplevel;
	connection_bar_struct *cb;
	dspview_struct *dv;
	yc_core_struct *cp = YC_CORE(core_ptr);
	yconsole_struct *yc = (yconsole_struct *)g_malloc0(
	    sizeof(yconsole_struct)
	);
	if(yc == NULL)
	    return(NULL);


	/* Reset values. */
	yc->initialized = TRUE;
	yc->map_state = FALSE;
	yc->busy_count = 0;
	yc->accelgrp = accelgrp = gtk_accel_group_new();
	yc->core_ptr = core_ptr;

	yc->con = NULL;
	yc->yshm_buf = NULL;
	yc->yshm_buf_len = 0;
	yc->yshm_id = -1;

	yc->start_time = 0;
	yc->audio_mode_name = NULL;
	yc->sample_rate = 0;
	yc->sample_size = 0;
	yc->channels = 0;
	yc->cycle_us = 0;
	yc->cycle_load = 0.0f;
	yc->num_fragments = 0;
	yc->fragment_size = 0;
	yc->flip_stereo = FALSE;
	yc->shellout = FALSE;
	yc->version_major = yc->version_minor = 0;
	yc->server_name = yc->sound_name = NULL;


	/* Create toplevel. */
	yc->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_name(w, "yconsole window");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_CONSOLE_SAVE_POSITIONS))
	    {
		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_CONSOLE_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_CONSOLE_Y)
		);
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_yconsole_48x48_xpm);
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(YConsoleDeleteEventCB), yc
	);

	/* Create connection bar. */
	yc->conbar = cb = ConBarNew(
	    CONBAR_DISPLAY_FLAG_WINDOW_MENU | CONBAR_DISPLAY_FLAG_LOCATION |
	    CONBAR_DISPLAY_FLAG_DSPVIEW | CONBAR_DISPLAY_FLAG_AUDIO |
	    CONBAR_DISPLAY_FLAG_REFRESH | CONBAR_DISPLAY_FLAG_SHELLOUT |
	    CONBAR_DISPLAY_FLAG_SHUTDOWN,
	    cp->dsp_bg_img_width + 2, cp->dsp_bg_img_height + 2,
	    yc,
	    YConsoleConnectCB, YConsoleYServerListCB, YConsoleCloseCB,
	    YConsoleNewYConsoleCB, YConsoleNewYAudioCB, YConsoleNewYMixerCB,
	    YConsoleNewYPlayerCB, YConsoleUptionsCB,
	    YConsoleRefreshCB, YConsoleShelloutCB, YConsoleShutdownCB,
	    toplevel, accelgrp
	);
	w = CONBAR_AUDIO(cb);
	gtk_widget_add_events(
	    w, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YConsoleConBarAudioButtonCB), yc
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YConsoleConBarAudioButtonCB), yc
	);
	dv = ConBarGetDSPView(cb);
	ConBarMap(cb);

	if(cp != NULL)
	    DSPViewSetBG(
		dv, cp->dsp_bg_img,
		cp->dsp_bg_img_width, cp->dsp_bg_img_height,
		cp->dsp_bg_img_bpl, 4
	    );
	DSPViewSetDisplayStyle(dv, DSPVIEW_DISPLAY_STYLE_NOVA);
	DSPViewSetMergeMode(dv, DSPVIEW_MERGE_MODE_SET);
	DSPViewSetDSPColor(dv, DSPVIEW_DEF_DSP_COLOR);  /* 0xAABBGGRR. */
	w = DSPVIEW_VIEW(dv);
	gtk_widget_add_events(
	    w, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YConsoleDSPViewButtonCB), yc
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YConsoleDSPViewButtonCB), yc
	);


	/* Update title. */
	YConsoleSetTitle(yc, NULL);

	return(yc);
}

/*
 *	Update menus on yconsole.
 */
void YConsoleUpdateMenus(yconsole_struct *yc)
{

}

/*
 *	Mark yconsole as busy or not.
 */
void YConsoleSetBusy(yconsole_struct *yc, gboolean is_busy)
{
	GdkCursor *cursor;
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(yc == NULL)
	    return;

	core_ptr = YC_CORE(yc->core_ptr);
	if(core_ptr == NULL)
	    return;

	w = YCONSOLE_TOPLEVEL(yc);
	if(w != NULL)
	{
	    if(is_busy)
	    {
		/* Increase busy count. */
		yc->busy_count++;

		/* If already busy then don't change anything. */
		if(yc->busy_count > 1)
		    return;

		cursor = core_ptr->cursor_busy;
	    }
	    else
	    {
		/* Reduce busy count. */
		yc->busy_count--;
		if(yc->busy_count < 0)
		    yc->busy_count = 0;

		/* If still busy do not change anything. */
		if(yc->busy_count > 0)
		    return;

		cursor = NULL;  /* Use default cursor. */
	    }

	    /* Update toplevel window's cursor. */
	    if(w->window != NULL)
	    {
		gdk_window_set_cursor(w->window, cursor);
		gdk_flush();
	    }
	}
}

/*
 *	Map yconsole.
 */
void YConsoleMap(yconsole_struct *yc)
{
	if(yc == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YCONSOLE_TOPLEVEL(yc);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    yc->map_state = TRUE;
	}
}

/*
 *	Unmap yconsole.
 */
void YConsoleUnmap(yconsole_struct *yc)
{
	if(yc == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YCONSOLE_TOPLEVEL(yc);
	    if(w != NULL)
		gtk_widget_hide(w);
	    yc->map_state = FALSE;
	}
}

/*
 *	Delete the yconsole.
 */
void YConsoleDelete(yconsole_struct *yc)
{
	GtkWidget **w;


	if(yc == NULL)
	    return;

	if(yc->initialized)
	{
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }

	    /* Disconnect from Y server as needed, this will reset
	     * all connection values.
	     */
	    YConsoleDoDisconnect(yc, FALSE);

	    /* Delete connection bar. */
	    ConBarDelete(YCONSOLE_CONBAR(yc));
	    yc->conbar = NULL;

	    /* Destroy widgets. */
	    w = &yc->toplevel;
	    DO_DESTROY_WIDGET

	    if(yc->accelgrp != NULL)
	    {
		gtk_accel_group_unref(yc->accelgrp);
		yc->accelgrp = NULL;
	    }

	}

	/* Deallocate structure itself. */
	g_free(yc);
}
