#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/fio.h"

#include "guiutils.h"

#include "cfg.h"
#include "yserverlist.h"
#include "yserverlistcb.h"
#include "yc.h"
#include "yccfglist.h"
#include "config.h"

#include "images/icon_y_48x48.xpm"

#include "images/icon_yserver_location_20x20.xpm"
#include "images/icon_run_20x20.xpm"

#include "images/icon_ok_20x20.xpm"
#include "images/icon_add_20x20.xpm"
#include "images/icon_edit_20x20.xpm"
#include "images/icon_remove_20x20.xpm"
#include "images/icon_close_20x20.xpm"


void YServerListFileOpen(yserver_list_struct *sl, const gchar *path);
void YServerListFileSave(yserver_list_struct *sl, const gchar *path);

gint YServerListAddItem(
	yserver_list_struct *sl, gint row,
	const gchar *name,
	const gchar *address,
	gboolean is_local,
	const gchar *start_command,
	const gchar *comments
);
void YServerListUpdateItem(yserver_list_struct *sl, gint row);
void YServerListSavePosition(yserver_list_struct *sl);

yserver_list_struct *YServerListNew(gpointer core_ptr);
void YServerListUpdateMenus(yserver_list_struct *sl);
void YServerListMap(yserver_list_struct *sl);
void YServerListUnmap(yserver_list_struct *sl);
void YServerListDelete(yserver_list_struct *sl);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Loads all server list items from the file specified by path.
 */
void YServerListFileOpen(yserver_list_struct *sl, const gchar *path)
{
	GtkCList *clist;
	FILE *fp;
	gchar *buf, *line_ptr, **strv;

	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	/* Remove existing items from clist */
	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);
	gtk_clist_thaw(clist);

	/* Open the Y Server list file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return;

	/* Begin reading the file */
	gtk_clist_freeze(clist);
	buf = NULL;
	while(TRUE)
	{
	    /* Delete the previous line and get the next line */
	    g_free(buf);
	    buf = FGetStringLiteral(fp);
	    if(buf == NULL)
		break;

	    /* Get pointer to new line, seek past spaces and check if
	     * it is a comment.
	     */
	    line_ptr = buf;
	    while(ISBLANK(*line_ptr))
		line_ptr++;
	    /* Comment? */
	    if(*line_ptr == '#')
		continue;
	    /* Empty line? */
	    if(*line_ptr == '\0')
		continue;

	    /* Explode this line into strings with the deliminator ';' */
 	    strv = g_strsplit(line_ptr, ";", -1);
	    if(strv != NULL)
	    {
		gint strc = 0;
		while(strv[strc] != NULL)
		    strc++;

		/* Append a new row to the servers list */
		YServerListAddItem(
		    sl, -1,
		    (strc > 0) ? strv[0] : NULL,
		    (strc > 1) ? strv[1] : NULL,
		    (strc > 2) ?
			((atoi(strv[2])) ? TRUE : FALSE) : FALSE,
		    (strc > 3) ? strv[3] : NULL,
		    (strc > 4) ? strv[4] : NULL
		);

		g_strfreev(strv);
	    }
	}

	gtk_clist_thaw(clist);

	/* Close the Y Server list file */
	fclose(fp);

	YServerListUpdateMenus(sl);
}

/*
 *	Saves all server list items to the file specified by path.
 */
void YServerListFileSave(yserver_list_struct *sl, const gchar *path)
{
	GtkCList *clist;
	FILE *fp;
	gint i;
	const yserver_list_item_struct *item;

	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	/* Open the Y Server list file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	    return;

	/* Begin writing file */
	for(i = 0; i < clist->rows; i++)
	{
	    item = YSERVER_LIST_ITEM(
		gtk_clist_get_row_data(clist, i)
	    );
	    if(item == NULL)
		continue;

	    fprintf(
		fp,
		"%s;%s;%c;%s;%s\n",
		item->name,
		item->address,
		(item->is_local) ? '1' : '0',
		item->start_command,
		item->comments
	    );
	}

	/* Close the Y Server list file */
	fclose(fp);
}

/*
 *	Insert a new item at the specified row, if row is -1 then the
 *	item will be appended.
 *
 *	Returns the new row index or -1 on error.
 */
gint YServerListAddItem(
	yserver_list_struct *sl, gint row,
	const gchar *name,
	const gchar *address,
	gboolean is_local,
	const gchar *start_command,
	const gchar *comments
)
{
	gchar *v[3];
	GtkCList *clist;
	yserver_list_item_struct *item;


	if(sl == NULL)
	    return(-1);

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return(-1);

	/* Set up initial row values and append/insert new row */
	v[0] = "";
	v[1] = "";
	v[2] = "";
	if(row < 0)
	    row = gtk_clist_append(clist, v);
	else
	    row = gtk_clist_insert(clist, row, v);
	if(row < 0)
	    return(-1);

	/* Allocate item structure */
	item = (yserver_list_item_struct *)g_malloc0(
	    sizeof(yserver_list_item_struct)
	);
	if(item != NULL)
	{
	    /* Copy given valus to item */
	    item->name = (name != NULL) ? g_strdup(name) : NULL;
	    item->address = (address != NULL) ?
		g_strdup(address) : NULL;
	    item->is_local = is_local;
	    item->start_command = (start_command != NULL) ? 
		g_strdup(start_command) : NULL;
	    item->comments = (comments != NULL) ?
		g_strdup(comments) : NULL; 

	    /* Set data and destroy callback for this row */
	    gtk_clist_set_row_data_full(
		clist, row, item,
		(GtkDestroyNotify)YServerListItemDestroyCB
	    );
	}

	/* Update clist cell values with respect to the newly set values
	 * on the item structure.
	 */
	YServerListUpdateItem(sl, row);

	return(row);
}

/*
 *	Updates the row with its item data values.
 */
void YServerListUpdateItem(yserver_list_struct *sl, gint row)
{
	GtkCList *clist;
	yserver_list_item_struct *item;


	if(sl == NULL)
	    return;

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	if((row < 0) || (row >= clist->rows))
	    return;

	item = YSERVER_LIST_ITEM(
	    gtk_clist_get_row_data(clist, row)
	);
	if(item == NULL)
	    return;

	/* Name */
	if(item->name != NULL)
	    gtk_clist_set_pixtext(
		clist, row, 0,
		item->name, YC_LIST_PIXMAP_TEXT_SPACING,
		sl->pixmap_yserver, sl->mask_yserver
	    );

	/* Address */
	if(item->address != NULL)
	    gtk_clist_set_text(
		clist, row, 1, item->address
	    );

	/* Startable */
	if((item->start_command != NULL) ?
	    (*item->start_command != '\0') : FALSE
	)
	    gtk_clist_set_pixmap(
		clist, row, 2,
		sl->pixmap_yserver_start, sl->mask_yserver_start
	    );
	else
	    gtk_clist_set_text(
		clist, row, 2, ""
	    );
}


/*
 *      Save current position value of window to configuration list on
 *      core structure.
 */
void YServerListSavePosition(yserver_list_struct *sl)
{
	GtkWidget *w;
	yc_core_struct *core_ptr;


	if(sl == NULL)
	    return;

	core_ptr = YC_CORE(sl->core_ptr);
	if(core_ptr == NULL)
	    return;

	/* Do not save positions? */
	if(!YC_GET_CFG_I(
	    core_ptr, YC_CFG_PARM_SERVER_LIST_SAVE_POSITIONS
	))
	    return;

	/* Save toplevel position */
	w = YSERVER_LIST_TOPLEVEL(sl);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_SERVER_LIST_X, x);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_SERVER_LIST_Y, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_SERVER_LIST_WIDTH, width);
	    YC_SET_CFG_I(core_ptr, YC_CFG_PARM_SERVER_LIST_HEIGHT, height);
	}

	/* Save servers list column widths */
	w = YSERVER_LIST_SERVERS_CLIST(sl);
	if(w != NULL)
	{
	    gint i;
	    GList *glist = NULL;
	    GtkCList *clist = GTK_CLIST(w);
	    const gint columns = clist->columns;
	    GtkCListColumn *column;
	    cfg_intlist_struct *intlist;
	    for(i = 0; i < columns; i++)
	    {
		column = &clist->column[i];
		glist = g_list_append(
		    glist, (gpointer)column->width
		);
	    }
	    intlist = CFGIntListNew(glist);
	    g_list_free(glist);
	    YC_SET_CFG_INTLIST(
		core_ptr, YC_CFG_PARM_SERVER_LIST_COLUMN_WIDTHS,
		intlist
	    );
	    CFGIntListDelete(intlist);
	}

}


/*
 *	Create a new yserver list.
 */
yserver_list_struct *YServerListNew(gpointer core_ptr)
{
	gint i, border_major = 5, border_minor = 2;
	gchar *heading[3];
	gint list_column_width[3];
	GdkWindow *window;
	GtkStyle *style;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2, *parent3, *parent4,
		*toplevel, *main_vbox, *menu;
	GtkCList *clist;
	yc_core_struct *cp = YC_CORE(core_ptr);
	yserver_list_struct *sl = (yserver_list_struct *)g_malloc0(
	    sizeof(yserver_list_struct)
	);
	if(sl == NULL)
	    return(NULL);

	list_column_width[0] = 120;
	list_column_width[1] = 150;
	list_column_width[2] = 30;

	/* Reset values */
	sl->initialized = TRUE;
	sl->map_state = FALSE;
	sl->busy_count = 0;
	sl->accelgrp = accelgrp = gtk_accel_group_new();
	sl->core_ptr = core_ptr;

	sl->rel_toplevel = NULL;

	/* Create toplevel */
	sl->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name(w, "yserver list window");
	if(cp != NULL)
	{
	    if(YC_GET_CFG_I(cp, YC_CFG_PARM_SERVER_LIST_SAVE_POSITIONS))
	    {
		const cfg_intlist_struct *intlist;

		gtk_widget_set_uposition(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_SERVER_LIST_X),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_SERVER_LIST_Y)
		);
		gtk_widget_set_usize(
		    w,
		    YC_GET_CFG_I(cp, YC_CFG_PARM_SERVER_LIST_WIDTH),
		    YC_GET_CFG_I(cp, YC_CFG_PARM_SERVER_LIST_HEIGHT)
		);

		intlist = YC_GET_CFG_INTLIST(
		    cp, YC_CFG_PARM_SERVER_LIST_COLUMN_WIDTHS
		);
		if(intlist != NULL)
		{
		    GList *glist = intlist->list;
		    for(i = 0; i < 4; i++)
		    {
			if(glist == NULL)
			    break;

			list_column_width[i] = (gint)glist->data;
			glist = g_list_next(glist);
		    }
		}
	    }
	}
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(w), "Y Servers List");
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_RESIZE |
		GDK_FUNC_MINIMIZE | GDK_FUNC_MAXIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_y_48x48_xpm);
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(YServerListDeleteEventCB), sl
	);
	parent = w;


	/* Main vbox */
	sl->main_vbox = main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Scrolled window for clist */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Contents clist */
	heading[0] = "Name";
	heading[1] = "Address";
	heading[2] = "";
	sl->servers_clist = w = gtk_clist_new_with_titles(3, heading);
	clist = GTK_CLIST(w);
	style = gtk_widget_get_style(w);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_set_column_width(clist, 0, list_column_width[0]);
	gtk_clist_set_column_width(clist, 1, list_column_width[1]);
	gtk_clist_set_column_width(clist, 2, list_column_width[2]);
	gtk_clist_set_row_height(clist, YC_LIST_ROW_SPACING);
	gtk_widget_add_events(w, GDK_BUTTON_PRESS_MASK);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(YServerListKeyEventCB), sl
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(YServerListKeyEventCB), sl
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(YServerListButtonEventCB), sl
	);
	gtk_signal_connect_after(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(YServerListButtonEventCB), sl
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "select_row",
	    GTK_SIGNAL_FUNC(YServerListSelectRowCB), sl
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "unselect_row",
	    GTK_SIGNAL_FUNC(YServerListUnselectRowCB), sl
	);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);


	/* Load pixmap and mask pairs for servers list items */
	sl->pixmap_yserver = gdk_pixmap_create_from_xpm_d(
	    window, &sl->mask_yserver,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon_yserver_location_20x20_xpm
	);
	sl->pixmap_yserver_start = gdk_pixmap_create_from_xpm_d(
	    window, &sl->mask_yserver_start,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon_run_20x20_xpm
	);


	/* Hbox for buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Add button */
	sl->add_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_add_20x20_xpm, "Add...", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListAddCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, 'a', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);

	/* Edit button */
	sl->edit_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_edit_20x20_xpm, "Edit...", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListEditCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, 'e', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Return, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);

	/* Remove button */
	sl->remove_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_remove_20x20_xpm, "Remove", NULL
	);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH, GUI_BUTTON_HLABEL_HEIGHT
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListRemoveCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, 'r', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Delete, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);


	w = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Up button */
	sl->shift_up_btn = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListShiftUpCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Up, GDK_CONTROL_MASK | GDK_MOD1_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);
	parent4 = w;
	/* Arrow */
	w = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);
	gtk_widget_set_usize(w, 20, 20);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);

	/* Down button */
	sl->shift_down_btn = w = gtk_button_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListShiftDownCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Down, GDK_CONTROL_MASK | GDK_MOD1_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);
	parent4 = w;
	/* Arrow */
	w = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	gtk_widget_set_usize(w, 20, 20);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);


	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Hbox for buttons */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Connect button */
	sl->connect_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_ok_20x20_xpm, "Connect", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListConnectCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);

	/* Start button */
	sl->start_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_run_20x20_xpm, "Start", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListStartCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, 'r', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);


	/* Close button */
	sl->close_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_close_20x20_xpm, "Close", NULL
	);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_widget_set_usize(
	    w,
	    GUI_BUTTON_HLABEL_WIDTH_DEF, GUI_BUTTON_HLABEL_HEIGHT_DEF
	);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(YServerListCloseCB), sl
	);
	gtk_accel_group_add(
	    accelgrp, 'w', GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);


	/* Right click menu */
	sl->menu = menu = GUIMenuCreate();
	if(menu != NULL)
	{
	    guchar **icon;
	    const gchar *label;
	    gint accel_key = 0;
	    guint accel_mods = 0;
	    gpointer accel_group = NULL;
	    gpointer mclient_data = sl;
	    void (*func_cb)(GtkWidget *w, gpointer) = NULL;
	    gpointer fw;

#define DO_ADD_MENU_ITEM_LABEL	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accel_group, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP		\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}
	    icon = (guchar **)icon_ok_20x20_xpm;
	    label = "Connect";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListConnectCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->connect_mi = w;

	    icon = (guchar **)icon_run_20x20_xpm;
	    label = "Start Server";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListStartCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->start_mi = w;

	    DO_ADD_MENU_SEP

	    icon = (guchar **)icon_add_20x20_xpm;
	    label = "Add...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListAddCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->add_mi = w;

	    icon = (guchar **)icon_edit_20x20_xpm;
	    label = "Edit...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListEditCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->edit_mi = w;

	    icon = (guchar **)icon_remove_20x20_xpm;
	    label = "Remove";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListRemoveCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->remove_mi = w;

	    DO_ADD_MENU_SEP

	    icon = NULL;
	    label = "Shift Up";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListShiftUpCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->shift_up_mi = w;

	    icon = NULL;
	    label = "Shift Down";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = YServerListShiftDownCB;
	    DO_ADD_MENU_ITEM_LABEL
	    sl->shift_down_mi = w;

	}


	YServerListUpdateMenus(sl);

	return(sl);
}

/*
 *	Update yserver list menus and widgets.
 */
void YServerListUpdateMenus(yserver_list_struct *sl)
{
	gboolean sensitive;
	gint sel_row;
	GList *glist;
	GtkWidget *w;
	GtkCList *clist;


	if(sl == NULL)
	    return;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) gtk_widget_set_sensitive(w, sensitive); }

	clist = (GtkCList *)YSERVER_LIST_SERVERS_CLIST(sl);
	if(clist == NULL)
	    return;

	/* Get last selected row on servers list */
	glist = clist->selection_end;
	sel_row = (glist != NULL) ? (gint)glist->data : -1;

	sensitive = (sel_row >= 0) ? TRUE : FALSE;

	w = sl->edit_btn;
	DO_SET_SENSITIVE
	w = sl->remove_btn;
	DO_SET_SENSITIVE
	w = sl->shift_up_btn;
	DO_SET_SENSITIVE
	w = sl->shift_down_btn;
	DO_SET_SENSITIVE
	w = sl->connect_btn;
	DO_SET_SENSITIVE
	w = sl->start_btn;
	DO_SET_SENSITIVE

	w = sl->connect_mi;
	DO_SET_SENSITIVE
	w = sl->start_mi;
	DO_SET_SENSITIVE
	w = sl->edit_mi;
	DO_SET_SENSITIVE
	w = sl->remove_mi;
	DO_SET_SENSITIVE
	w = sl->shift_up_mi;
	DO_SET_SENSITIVE
	w = sl->shift_down_mi;
	DO_SET_SENSITIVE

#undef DO_SET_SENSITIVE
}

/*
 *      Map yserver list.
 */
void YServerListMap(yserver_list_struct *sl)
{
	if(sl == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YSERVER_LIST_TOPLEVEL(sl);
	    if(w != NULL)
		gtk_widget_show_raise(w);
	    sl->map_state = TRUE;
	}
}


/*
 *	Unmap yserver list.
 */
void YServerListUnmap(yserver_list_struct *sl)
{
	if(sl == NULL)
	    return;

	if(TRUE)
	{
	    GtkWidget *w = YSERVER_LIST_TOPLEVEL(sl);
	    if(w != NULL)
		gtk_widget_hide(w);
	    sl->map_state = FALSE;
	}
}

/*
 *      Delete the yserver list.
 */
void YServerListDelete(yserver_list_struct *sl)
{
	GdkBitmap **mask;
	GdkPixmap **pixmap;
	GtkWidget **w;


	if(sl == NULL)
	    return;

	if(sl->initialized)
	{
#define DO_UNREF_PIXMAP_MASK	\
{ if(*pixmap != NULL) \
   { gdk_pixmap_unref(*pixmap); *pixmap = NULL; } \
  if(*mask != NULL) \
   { gdk_bitmap_unref(*mask); *mask = NULL; } \
}
 
#define DO_DESTROY_WIDGET       \
{ if(*w != NULL) \
   { gtk_widget_destroy(*w); *w = NULL; } \
}

	    /* Destroy widgets */
	    w = &sl->menu;
	    sl->connect_mi = NULL;
	    sl->start_mi = NULL;
	    sl->add_mi = NULL;
	    sl->edit_mi = NULL;
	    sl->remove_mi = NULL;
	    sl->shift_up_mi = NULL;
	    sl->shift_down_mi = NULL;
	    DO_DESTROY_WIDGET

	    w = &sl->add_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->edit_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->remove_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->shift_up_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->shift_down_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->connect_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->start_btn;
	    DO_DESTROY_WIDGET
	    w = &sl->close_btn;
	    DO_DESTROY_WIDGET

	    w = &sl->servers_clist;
	    DO_DESTROY_WIDGET

	    w = &sl->main_vbox;
	    DO_DESTROY_WIDGET
	    w = &sl->toplevel;
	    DO_DESTROY_WIDGET

	    if(sl->accelgrp != NULL)
	    {
		gtk_accel_group_unref(sl->accelgrp);
		sl->accelgrp = NULL;
	    }

	    pixmap = &sl->pixmap_yserver;
	    mask = &sl->mask_yserver;
	    DO_UNREF_PIXMAP_MASK
	    pixmap = &sl->pixmap_yserver_start;
	    mask = &sl->mask_yserver_start;
	    DO_UNREF_PIXMAP_MASK

#undef DO_DESTROY_WIDGET
#undef DO_UNREF_PIXMAP_MASK
	}

	/* Deallocate structure itself */
	g_free(sl);
}
